/*!
    \file  colormap.c
    \brief Colormap support
*/

/*
 * Copyright 2005 the OpenGLEAN Project.
 *
 * Portions copyright (C) 2004, the OpenGLUT project contributors.
 * OpenGLUT branched from freeglut in February, 2004.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#if TARGET_HOST_UNIX_X11
/*
 * Helper function for ensuring that the current window has
 * its color map.  Currently only for X.  If WIN32 has a similar
 * concept of shared and private color maps, this will need a
 * WIN32 branch.
 */
static void oghEnsureColormap (void)
{
    if (!ogStructure.Window->Window.Colormap)
    {
        ogStructure.Window->Window.Colormap = XCreateColormap (
            ogDisplay.Display,
            ogStructure.Window->Window.Handle,
            ogStructure.Window->Window.VisualInfo->visual,
            AllocAll
        );
        if (!ogStructure.Window->Window.Colormap)
            ogError (
                "glutSetColor (): Could not allocate colormap for window."
            );
        XSetWindowColormap (
            ogDisplay.Display,
            ogStructure.Window->Window.Handle,
            ogStructure.Window->Window.Colormap
        );
    }
}
static Colormap oghGetWindowColormap (SOG_Window * w)
{
    Colormap ret = w->Window.Colormap;

    if (!ret)
        ret = DefaultColormap (
            ogDisplay.Display, DefaultScreen (ogDisplay.Display)
        );
    return ret;
}

#endif
/*!
    \fn
    \brief    Sets an indexed color-mode entry.
    \ingroup  colormap
    \param    nColor    The palette entry to change.
    \param    red       New red value for palette entry.
    \param    green     New green value for palette entry.
    \param    blue      New blue value for palette entry.

              glutSetCursor() allows you to set individual color-map entries
              in a \a GLUT_INDEX type of display.
              Uses the <i>current window</i>.
              Respects the current overlay setting.

              The minimum and maximum range for \a nColor is not
              well defined at this time.

              \a red, \a green, and \a blue are values as for
              glColor3f().

              It is a fatal error either to call this function without
              properly initializing OpenGLEAN via glutInit() or to
              call this function without a <i>current window</i>.

    \note     Obviously should do nothing for non-colormapped windows.

    \see      glutGetColor(), glutCopyColorMap()

    \todo     Error checking.

    \todo     Testing.

    \todo     WIN32 version.

    \bug      Probably should clone the root window's color map to start.
*/
void OGAPIENTRY glutSetColor (
    int nColor, GLfloat red, GLfloat green, GLfloat blue
)
{
    OPENGLEAN_REQUIRE_READY ("glutSetColor");
    OPENGLEAN_REQUIRE_WINDOW ("glutSetColor");
#if TARGET_HOST_UNIX_X11
    oghEnsureColormap ();
    if (0 > red)
        red = 0;
    else if (1 < red)
        red = 1;
    if (0 > green)
        green = 0;
    else if (1 < green)
        green = 1;
    if (0 > blue)
        blue = 0;
    else if (1 < blue)
        blue = 1;
    {
        XColor          color;

        color.pixel = (unsigned long) nColor;
        color.red   = (unsigned short) (65535 * red);
        color.green = (unsigned short) (65535 * green);
        color.blue  = (unsigned short) (65335 * blue);
        color.flags = DoRed | DoGreen | DoBlue;
        XStoreColor (
            ogDisplay.Display, ogStructure.Window->Window.Colormap, &color
        );
    }
#else
    ogWarning (
        "glutSetColor (%d, %f, %f, %f): Unimplemented.",
        nColor, red, green, blue
    );
#endif
}

/*!
    \fn
    \brief    Gets an indexed color-mode entry's Red, Green, or Blue value.
    \ingroup  colormap
    \param    color     The palette entry to fetch.
    \param    component Whether to fetch Red, Green, or Blue.

              Allows you to get individual color-map entries
              in a \a GLUT_INDEX type of display.
              Respects the current layer setting.
              Works on the <i>current window</i>.

              \a component may be any of:

               - \a GLUT_RED
               - \a GLUT_GREEN
               - \a GLUT_BLUE

              It is a fatal error either to call this function without
              properly initializing OpenGLEAN via glutInit() or to
              call this function without a <i>current window</i>.

    \note     Obviously should do nothing for non-colormapped windows.

    \see      glutSetColor(), glutCopyColorMap()

    \todo     Untested.

    \todo     WIN32 version.
*/
GLfloat OGAPIENTRY glutGetColor (int color, int component)
{
    GLfloat ret = 0;

    OPENGLEAN_REQUIRE_READY ("glutGetColor");
    OPENGLEAN_REQUIRE_WINDOW ("glutGetColor");
#if TARGET_HOST_UNIX_X11
    {
        XColor xcolor;
        Colormap cmap = oghGetWindowColormap (ogStructure.Window);

        xcolor.pixel = color;
        xcolor.flags = DoRed | DoGreen | DoBlue;
        if (!XQueryColor (ogDisplay.Display, cmap, &xcolor))
            ogError (
                "glutGetColor (%d, %d): XQueryColor failed.",
                color, component
            );
        switch (component)
        {
        case GLUT_RED:
            ret = xcolor.red;
            break;
        case GLUT_GREEN:
            ret = xcolor.green;
            break;
        case GLUT_BLUE:
            ret = xcolor.blue;
            break;
        default:
            ogError (
                "glutGetColor (%d, %d): Unrecognized component, %d",
                color, component,
                component
            );
            break;
        }
    }
#else
    ogWarning ("glutSetColor (%d, %d): Unimplemented.", color, component);
#endif
    return ret;
}

/*!
    \fn
    \brief    Copies a color map between windows.
    \ingroup  colormap
    \param    window    The window to copy <i>to</i>

              Allows you to copy an entire color map from
              one window to another.  This function copies
              <b>from</b> the <i>current window</i>.  It
              copies <b>to</b> the indicated \a window.
              Respects the current layer setting.

              It does <b>not</b> set the <i>current window</i>

              It is a fatal error either to call this function without
              properly initializing OpenGLEAN via glutInit() or to
              call this function without a <i>current window</i>.

    \note     There is no GLUT API for copying
              a color map into a cache, so if you (e.g.)
              want to change lots of colors very quickly,
              you either need to loop over glutSetColor(),
              or create a color-mapped window on the side.

    \note     Obviously should do nothing for non-colormapped windows.

    \see      glutSetColor(), glutGetColor()

    \todo     Determine if we <b>should</b> set the <i>current window</i>

    \todo     Test.

    \todo     WIN32 version.
*/
void OGAPIENTRY glutCopyColormap (int window)
{
    OPENGLEAN_REQUIRE_READY ("glutCopyColormap");
    OPENGLEAN_REQUIRE_WINDOW ("glutCopyColormap");
#if TARGET_HOST_UNIX_X11
    int cmap_size  = glutGet (GLUT_WINDOW_COLORMAP_SIZE);
    XColor *colors = malloc (cmap_size * sizeof (XColor));

    if (!colors)
        ogError (
            "glutCopyColormap (%d): Could not allocate XColors array.",
            window
        );
    int  success = XQueryColors (
        ogDisplay.Display, oghGetWindowColormap (ogStructure.Window),
        colors, cmap_size
    );

    success = XQueryColors (
        ogDisplay.Display, oghGetWindowColormap (ogStructure.Window),
        colors, cmap_size
    );
    if (!success)
        ogError (
            "glutCopyColormap (%d): Could not read colors.", window
        );
    success = XStoreColors (
        ogDisplay.Display, oghGetWindowColormap (ogWindowByID (window)),
        colors, cmap_size
    );
    if (!success)
        ogError (
            "glutCopyColormap (%d): Could not write colors.", window
        );
    free (colors);
#else
    ogWarning ("glutCopyColormap (%d): Unimplemented.", window);
#endif
}
