/*!
    \file  display.c
    \brief Display message posting, context buffer swapping.
*/

/*
 * Copyright 2004, 2005, the OpenGLEAN Project.
 *
 * Portions copyright (C) 2004, the OpenGLUT project contributors.
 * OpenGLUT branched from freeglut in February, 2004.
 *
 * Copyright (c) 1999-2000 Pawel W. Olszta. All Rights Reserved.
 * Written by Pawel W. Olszta, <olszta@sourceforge.net>
 * Creation date: Fri Dec 3 1999
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/* -- INTERFACE FUNCTIONS -------------------------------------------------- */

/*!
    \fn
    \brief    Mark the current window as needing a redisplay.
    \ingroup  window

              Whenever circumstances indicate that your window is
              in need of being redisplayed, you may call glutPostRedisplay()
              to tell OpenGLEAN that you want to redraw your graphics.
              Multiple calls to this function may be coalesced by OpenGLEAN
              to avoid excessive invocation of your drawing support.

              The ultimate effect of this function is to call your
              Display callback for the <i>current window</i>.

              It is a fatal error either to call this function without
              properly initializing OpenGLEAN via glutInit() or to
              call this function without a <i>current window</i>.

    \see      glutPostWindowRedisplay(), glutPostOverlayRedisplay(),
              glutPostWindowOverlayRedisplay(), glutSwapBuffers(),
              glutDisplayFunc()
*/
void OGAPIENTRY glutPostRedisplay (void)
{
    OPENGLEAN_REQUIRE_READY ("glutPostRedisplay");
    OPENGLEAN_REQUIRE_WINDOW ("glutPostRedisplay");
    ogStructure.Window->State.Redisplay = GL_TRUE;
}

/*!
    \fn
    \brief    Swaps the buffers for the current window.
    \ingroup  window

              This function signals to OpenGLEAN that you are done drawing
              to the <i>current window</i> for now.  If your window is
              double-buffered (\a GLUT_DOUBLE param to glutInitDisplayMode()),
              then OpenGLEAN will swap the front buffer with the back buffer.

              This also computes your current frame-rate and prints the result
              on \a stderr if indicated by the \a GLUT_FPS environment
              variable.  The computed value is not necessarily the
              total frame rate, if you have multiple windows, as the
              statistic is the total number of buffer-swaps for the
              entire program.

              It is a fatal error either to call this function without
              properly initializing OpenGLEAN via glutInit() or to
              call this function without a <i>current window</i>.

    \note     This function has little or no effect for \a GLUT_SINGLE.

    \note     Frame rate is now calculated for <b>all</b> windows if
              you call glutSwapBuffers() on them.

    \see      glutPostRedisplay(), glutPostOverlayRedisplay(),
              glutPostWindowRedisplay(), glutPostWindowOverlayRedisplay(),
              glutInitDisplaymode()

    \todo     How does this interact with overlays?

    \todo     Consider making \a GLUT_FPS keep per-window stats in a
              multi-window program.
*/
void OGAPIENTRY glutSwapBuffers (void)
{
    OPENGLEAN_REQUIRE_READY ("glutSwapbuffers");
    OPENGLEAN_REQUIRE_WINDOW ("glutSwapbuffers");

#if TARGET_HOST_WIN32
    /*
     * XXX GLX does this when swapping buffers.
     * XXX Maybe single-buffered requires it?  WIN32 may also
     * XXX Require it in general.
     * XXX A "good" GLX implementation should not care either
     * XXX way, but a sloppy GLX may assume that the client
     * XXX never "does something stupid", so potentially this
     * XXX could have a significant slowdown on some GLX.
     */
    glFlush ();
#endif
    if (!ogStructure.Window->Window.DoubleBuffered)
        ; //XXX Maybe make it a glFlush () ?
    else
    {
#if TARGET_HOST_UNIX_X11
        glXSwapBuffers (ogDisplay.Display, ogStructure.Window->Window.Handle);
#elif TARGET_HOST_WIN32 || TARGET_HOST_WINCE
        if (!(ogStructure.Window->State.IsOffscreen))
            SwapBuffers (ogStructure.Window->Window.Device);
#endif
    }
    /* GLUT_FPS env var support */
    if (ogState.FPSInterval)
    {
        GLint t = glutGet (GLUT_ELAPSED_TIME);

        ogState.SwapCount++;
        if (ogState.SwapTime == 0)
            ogState.SwapTime = t;
        else if (t - ogState.SwapTime > ogState.FPSInterval)
        {
            float time = .001f * (t - ogState.SwapTime);
            float fps = (float) ogState.SwapCount / time;

            fprintf (
                stderr,
                "OpenGLEAN: %d frames in %.2f seconds = %.2f FPS\n",
                ogState.SwapCount, time, fps
            );
            ogState.SwapTime = t;
            ogState.SwapCount = 0;
        }
    }
}

/*!
    \fn
    \brief    Mark an indicated window as needing a redisplay.
    \ingroup  window
    \param    windowID    The OpenGLEAN window id to be affected.

              Similar to glutPostRedisplay(), except that instead
              of affecting the <i>current window</i>, this function
              affects an arbitrary window, indicated by the
              \a windowID parameter.

              It is a fatal error to call this function without
              properly initializing OpenGLEAN via glutInit().

    \note     If the requested \a windowID cannot be mapped to a
              window, this silently does nothing.  Perhaps we
              should print a warning, or even exit with an error
              message?
    \see      glutPostRedisplay(), glutPostOverlayRedisplay(),
              glutPostWindowOverlayRedisplay(), glutSwapBuffers(),
              glutDisplayFunc(), glutCreateWindow(), glutCreateSubWindow()
*/
void OGAPIENTRY glutPostWindowRedisplay (int windowID)
{
    SOG_Window *window;

    OPENGLEAN_REQUIRE_READY ("glutPostWindowRedisplay");
    window = ogWindowByID (windowID);
    if (window)
        window->State.Redisplay = GL_TRUE;
}
