/*!
    \file  ext.c
    \brief OpenGL extensions
*/

/*
 * Copyright 2004, 2005, the OpenGLEAN Project.
 *
 * Portions copyright (C) 2004, the OpenGLUT project contributors.
 * OpenGLUT branched from freeglut in February, 2004.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

struct name_address_pair
{
    const char *name;
    void *address; /* XXX Should be {const void *address}! */
};


/*
 * The freeglut code staticly declared an array to hold name->function
 * mappings.  Due to a compiler bug in MinGW, this causes problems when
 * trying to build a ".dll" shared object library.  There are a few
 * ways to respond to this:
 *
 *  1. Since it's a compiler bug, wait for the compiler to get fixed.
 *  2. Switch from a table method to an if...else if...else if...
 *     construct.
 *  3. Build the table dynamicly.
 *  4. Write the list of functions to another file, and add a support
 *     tool to OpenGLEAN which compiles that file into a state machine
 *     in C (if(('g'==*name++)&&('l'==*name++)&&...){switch(*name++)...})
 *
 * (In cases 3 and 4, a trivial variation is to use hashing.  I don't
 *  think that that's appropriate for our purposes, though.)
 *
 * OpenGLEAN opts for the dynamic table construction.  The state machine
 * approach is probably the ideal: It would be faster, avoid dynamic
 * table allocation/construction, and reduce redundancy (there is at
 * least one other spot that has a table of all public OpenGLEAN
 * functions).
 *
 * XXX If we keep this data dynamically allocated this way, we should
 * XXX probably track it in the global state, and clean it up on exit.
 * XXX However, I think that the state-machine/recognizer approach is
 * XXX the way to go...
 */
static struct name_address_pair *glut_functions = NULL;
static unsigned glut_functions_count = 0;
static unsigned glut_functions_space = 0;
static int oghNameAddressPairComparator (const void *_a, const void *_b)
{
    const struct name_address_pair *a = _a;
    const struct name_address_pair *b = _b;
    return strcmp (a->name, b->name);
}
static int oghAddGlutProc (const char *const name, void *const address)
{
    int ret = 0;

    if (!glut_functions)
    {
        glut_functions = malloc (8 * sizeof (*glut_functions));
        if (glut_functions)
            glut_functions_space = 8;
    }
    if (glut_functions)
    {
        if (glut_functions_count >= glut_functions_space)
        {
            struct name_address_pair *tmp;
            tmp = realloc (
                glut_functions,
                2 * glut_functions_space * sizeof (*glut_functions)
            );
            if (tmp)
            {
                glut_functions = tmp;
                glut_functions_space *= 2;
            }
        }
        if (glut_functions_space > glut_functions_count)
        {
            glut_functions [glut_functions_count].name = name;
            glut_functions [glut_functions_count].address = address;
            ++glut_functions_count;
            ret = 1;
        }
    }
    return ret;
}

static int oghGlutGetProcAddressInit (void)
{
    int ret = 1;
#define ENTRY(f) ret = ret && oghAddGlutProc (#f, (void *)f)
    ENTRY (glutInit);
    ENTRY (glutInitDisplayMode);
    ENTRY (glutInitDisplayString);
    ENTRY (glutInitWindowPosition);
    ENTRY (glutInitWindowSize);
    ENTRY (glutMainLoop);
    ENTRY (glutCreateWindow);
    ENTRY (glutCreateSubWindow);
    ENTRY (glutDestroyWindow);
    ENTRY (glutPostRedisplay);
    ENTRY (glutPostWindowRedisplay);
    ENTRY (glutSwapBuffers);
    ENTRY (glutGetWindow);
    ENTRY (glutSetWindow);
    ENTRY (glutSetWindowTitle);
    ENTRY (glutSetIconTitle);
    ENTRY (glutPositionWindow);
    ENTRY (glutReshapeWindow);
    ENTRY (glutPopWindow);
    ENTRY (glutPushWindow);
    ENTRY (glutIconifyWindow);
    ENTRY (glutShowWindow);
    ENTRY (glutHideWindow);
    ENTRY (glutFullScreen);
    ENTRY (glutSetCursor);
    ENTRY (glutWarpPointer);
    ENTRY (glutEstablishOverlay);
    ENTRY (glutRemoveOverlay);
    ENTRY (glutUseLayer);
    ENTRY (glutPostOverlayRedisplay);
    ENTRY (glutPostWindowOverlayRedisplay);
    ENTRY (glutShowOverlay);
    ENTRY (glutHideOverlay);
    ENTRY (glutCreateMenu);
    ENTRY (glutDestroyMenu);
    ENTRY (glutGetMenu);
    ENTRY (glutSetMenu);
    ENTRY (glutAddMenuEntry);
    ENTRY (glutAddSubMenu);
    ENTRY (glutChangeToMenuEntry);
    ENTRY (glutChangeToSubMenu);
    ENTRY (glutRemoveMenuItem);
    ENTRY (glutAttachMenu);
    ENTRY (glutDetachMenu);
    ENTRY (glutDisplayFunc);
    ENTRY (glutReshapeFunc);
    ENTRY (glutKeyboardFunc);
    ENTRY (glutMouseFunc);
    ENTRY (glutMotionFunc);
    ENTRY (glutPassiveMotionFunc);
    ENTRY (glutEntryFunc);
    ENTRY (glutVisibilityFunc);
    ENTRY (glutIdleFunc);
    ENTRY (glutTimerFunc);
    ENTRY (glutSpecialFunc);
    ENTRY (glutSpaceballMotionFunc);
    ENTRY (glutSpaceballRotateFunc);
    ENTRY (glutSpaceballButtonFunc);
    ENTRY (glutButtonBoxFunc);
    ENTRY (glutDialsFunc);
    ENTRY (glutTabletMotionFunc);
    ENTRY (glutTabletButtonFunc);
    ENTRY (glutMenuStatusFunc);
    ENTRY (glutOverlayDisplayFunc);
    ENTRY (glutWindowStatusFunc);
    ENTRY (glutKeyboardUpFunc);
    ENTRY (glutSpecialUpFunc);
    ENTRY (glutJoystickFunc);
    ENTRY (glutSetColor);
    ENTRY (glutGetColor);
    ENTRY (glutCopyColormap);
    ENTRY (glutGet);
    ENTRY (glutDeviceGet);
    ENTRY (glutExtensionSupported);
    ENTRY (glutGetModifiers);
    ENTRY (glutLayerGet);
    ENTRY (glutBitmapCharacter);
    ENTRY (glutBitmapWidth);
    ENTRY (glutStrokeCharacter);
    ENTRY (glutStrokeWidth);
    ENTRY (glutBitmapLength);
    ENTRY (glutStrokeLength);
    ENTRY (glutWireSphere);
    ENTRY (glutSolidSphere);
    ENTRY (glutWireCone);
    ENTRY (glutSolidCone);
    ENTRY (glutWireCube);
    ENTRY (glutSolidCube);
    ENTRY (glutWireTorus);
    ENTRY (glutSolidTorus);
    ENTRY (glutWireDodecahedron);
    ENTRY (glutSolidDodecahedron);
    ENTRY (glutWireTeapot);
    ENTRY (glutSolidTeapot);
    ENTRY (glutWireOctahedron);
    ENTRY (glutSolidOctahedron);
    ENTRY (glutWireTetrahedron);
    ENTRY (glutSolidTetrahedron);
    ENTRY (glutWireIcosahedron);
    ENTRY (glutSolidIcosahedron);
    ENTRY (glutVideoResizeGet);
    ENTRY (glutSetupVideoResizing);
    ENTRY (glutStopVideoResizing);
    ENTRY (glutVideoResize);
    ENTRY (glutVideoPan);
    ENTRY (glutReportErrors);
    ENTRY (glutIgnoreKeyRepeat);
    ENTRY (glutSetKeyRepeat);
    ENTRY (glutForceJoystickFunc);
#if !TARGET_HOST_WINCE /* XXX Should NOT be conditional! */
    ENTRY (glutGameModeString);
    ENTRY (glutEnterGameMode);
    ENTRY (glutLeaveGameMode);
    ENTRY (glutGameModeGet);
#endif
    /* freeglut extensions */
    ENTRY (glutMainLoopEvent);
    ENTRY (glutLeaveMainLoop);
    ENTRY (glutCloseFunc);
    ENTRY (glutWMCloseFunc);
    ENTRY (glutMenuDestroyFunc);
    ENTRY (glutSetOption);
    ENTRY (glutSetWindowData);
    ENTRY (glutGetWindowData);
    ENTRY (glutSetMenuData);
    ENTRY (glutGetMenuData);
    ENTRY (glutBitmapHeight);
    ENTRY (glutStrokeHeight);
    ENTRY (glutBitmapString);
    ENTRY (glutStrokeString);
    ENTRY (glutWireRhombicDodecahedron);
    ENTRY (glutSolidRhombicDodecahedron);
    ENTRY (glutWireSierpinskiSponge);
    ENTRY (glutSolidSierpinskiSponge);
    ENTRY (glutWireCylinder);
    ENTRY (glutSolidCylinder);
    ENTRY (glutGetProcAddress);
    ENTRY (glutMouseWheelFunc);
    /* OpenGLUT extensions */
    ENTRY (glutCreateMenuWindow);
    ENTRY (glutSetWindowStayOnTop);
    /* OpenGLEAN extensions */
    ENTRY (glutSolidCuboctahedron);
    ENTRY (glutWireCuboctahedron);
    /* Array terminator */
    /* { NULL, NULL } */
#undef ENTRY
    if (ret)
        qsort (
            glut_functions, glut_functions_count, sizeof (*glut_functions),
            oghNameAddressPairComparator
        );
    return ret;
}

/*!
    \fn
    \brief    Determine if an OpenGLEAN feature or extension is available.
    \ingroup  openglean
    \param    procName    Procedure name.

              Given a function name, searches for the function
              (or "procedure", hence "Proc") in
              internal tables.  If the function is found, a pointer
              to the function is returned.  If the function is not
              found, \a NULL is returned.

              In addition to an internal OpenGLEAN table, this
              function will also consult glX (on X systems) or
              wgl (on WIN32 and WINCE), if the OpenGLEAN tables
              do not have the requested function.  It should
              return any OpenGL, glX, or wgl function if those
              functions are available.

    \note     The code makes no effort towards efficiency.
              It is advised that you cache the pointer
              retrieved from this function if you need a
              certain function repeatedly.
    \see      glXGetProcAddress(), wglGetProcAddress()
*/
void *OGAPIENTRY glutGetProcAddress (const char *procName)
{
    static int need_init = 1;
    static int succ_init = 0;

    /* Try GLUT functions first */
    if (need_init)
    {
        need_init = 0;
        succ_init = oghGlutGetProcAddressInit ();
    }
    if (!succ_init)
        ogWarning ("glutGetProcAddress: Could not initialize table.");
    else
    {
        struct name_address_pair *pair;
        struct name_address_pair key;
        key.name = procName;
        pair = bsearch (
            &key,
            glut_functions, glut_functions_count, sizeof (*glut_functions),
            oghNameAddressPairComparator
        );
        if (pair)
            return pair->address;
    }

    /* Try core GL functions */
#if TARGET_HOST_WIN32 || TARGET_HOST_WINCE
    return (void *)wglGetProcAddress ((LPCSTR) procName);
#elif TARGET_HOST_UNIX_X11 && defined (GLX_ARB_get_proc_address)
    return (void *)glXGetProcAddressARB ((const GLubyte *) procName);
#else
    return NULL;
#endif
}
