/*!
    \file  misc.c
    \brief Miscellaneous functions
*/

/*
 * Copyright 2004, 2005 the OpenGLEAN Project.
 *
 * Portions copyright (C) 2004, the OpenGLUT project contributors.
 * OpenGLUT branched from freeglut in February, 2004.
 *
 * Copyright (c) 1999-2000 Pawel W. Olszta. All Rights Reserved.
 * Written by Pawel W. Olszta, <olszta@sourceforge.net>
 * Creation date: Thu Dec 9 1999
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/* -- INTERNAL FUNCTIONS --------------------------------------------------- */

/*
 * Similar to the de facto strdup(), but since strdup() is not
 * actually a standard C function, we provide our own.
 */
PRIVATE char *ogStrDup (const char *str)
{
    char *ret = malloc (strlen (str) + 1);
    char *dst = ret;

    if (ret)
        while (*dst++ = *str++)
            ;
    return ret;
}

/* -- INTERFACE FUNCTIONS -------------------------------------------------- */

/*!
    \fn
    \brief    Determine if an OpenGL extension is available.
    \ingroup  opengl
    \param    extension    A string-name of an extension.

              Returns 0 if OpenGLEAN cannot determine that the requested
              extension is definitely present.  Only checks extensions
              from glGetString().

              It is presently a fatal error to call this function without
              properly initializing OpenGLEAN via glutInit().
              If you have no <i>current window</i>, or if
              \a extenstion contains an ASCII SPACE character,
              or if \a extension cannot be found in
              glGetString(\a GL_EXTENSIONS),
              then 0 is returned.

              Else returns 1.

              It is a bit inconsistent that we quietly fail if there
              is no <i>current window</i>, but blow up with an error
              message if OpenGLEAN is not properly initialized.

    \see      glGetString()

    \todo     Consider an ogWarning() if there is no current rendering
              context.

    \todo     This is an ugly bit of code with 5 return statements,
              one of which is never reached.
*/
int OGAPIENTRY glutExtensionSupported (const char *extension)
{
    const char *extensions, *start;
    const size_t len = strlen (extension);

    OPENGLEAN_REQUIRE_READY ("glutExtensionSupported");

    /* Check for current window and thus a current context */
    if (!ogStructure.Window)
        return 0;

    if (strchr (extension, ' '))
        return 0;
    start = extensions = (const char *) glGetString (GL_EXTENSIONS);

    if (!extensions)
        return 0;

    while (1)
    {
        const char *p = strstr (extensions, extension);

        if (!p)
            return 0;

        /* Check that we matched at a word boundary */
        if ((p == start || ' ' == p [-1]) &&
                (' ' == p [len] || 0 == p [len]))
            return 1;
        /* skip the false match and continue */
        extensions = p + len;
    }

    return 0;
}

/*
 * XXX This should be cleaned up.
 * XXX  a) The library handle perhaps should be released when
 * XXX     the library deinitializes.
 * XXX  b) The function pointer, at least, should probably be
 * XXX     migrated to ogState.
 * XXX  c) Use a function-pointer in all cases, whether it's the GLU
 * XXX     function directly, the dlsym() result, or the "internal"
 * XXX     version.
 *
 * XXX The WIN32 parallel to dlopen()/dlsym() is:
 * XXX
 * XXX     #include <Windows.h>
 * XXX     HINSTANCE lib_handle = LoadLibrary( path );
 * XXX     func_ptr = ( cast )GetProcAddress( lib_handle, func_name );
 * XXX
 * XXX However, because WIN32 users rarely build programs for themselves
 * XXX nor use sophisticated package systems with dependencies, but
 * XXX rather rely upon each package to be precompiled and to take care
 * XXX of all installation dependencies, this may not be worth worrying
 * XXX  about.
 * XXX
 * XXX If we do this for WIN32, I'd think that implemented the standard
 * XXX UNIX calls in terms of the WIN32 calls, on WIN32, is better than
 * XXX doing inline WIN32 everywhere.
 */
#if DLOPEN_GLU
/*
 * This function pointer and #define are used to call the GLU
 * gluErrorString() function if we dynamically open it.
 */
static const GLubyte *(*_gluErrorString) (GLenum error);

#define gluErrorString _gluErrorString
/*
 * ogGLULibraryHandle points to the dlopen()ed GLU library.
 *
 * The two strings are used for finding the library and the
 * function in the library.
 *
 * {triedToDLOpen} is used to control whether we print an error
 * message on failure.  The first time we reach for GLU and can't
 * get it, we want a warning.  After that, we can skip the warning,
 * since we won't be saying anything new.
 */
static void *ogGLULibraryHandle;
static const char *ogGLUPath = "libGLU.so";
static int triedToDLOpen;

/*
 * og_gluErrorString() is a simplistic fall-back if we are unable
 * to dlopen()/dlsym() the real GLU function.
 */
static const GLubyte *og_gluErrorString (GLenum error)
{
    static char *ret;

    if (ret)
        free (ret);
    asprintf (&ret, "OpenGL/GLU error: %d", error);
    return (GLubyte *)ret;
}

#endif

/*!
    \fn
    \brief    Reports all available OpenGL errors.
    \ingroup  opengl

              Displays as an OpenGLEAN warning every OpenGL error
              that OpenGL remembers giving to us and which
              we have not processed.  Uses gluErrorString().

              This is forcibly done by OpenGLEAN periodically
              if \a -gldebug is one of the strings passed into
              glutInit() via \a argv.

    \note     On some systems, this can be conditionally compiled
              to explicitly open GLU as a dynamic library.
              There are advantages and disadvantages to this.
              Currently, resolving GLU is automatic.
              Currently, if the manual resolution is used, the
              GLU library is searched for without any special
              path or version information.

    \see      gluErrorString(), glutInit()
*/
void OGAPIENTRY glutReportErrors (void)
{
    GLenum error;

#if DLOPEN_GLU
    if (!ogGLULibraryHandle && !triedToDLOpen)
    {
        _gluErrorString = og_gluErrorString; /* "Default", safe setting */
        ogGLULibraryHandle = dlopen (ogGLUPath, DL_LAZY);
        if (!ogGLULibraryHandle)
            ogWarning (
                "Failed to locate library %s for GLU error strings.", ogGLUPath
            );
        else
        {
            _gluErrorString = dlsym (ogGLULibraryHandle, "gluErrorString");
            if (!_gluErrorString)
                ogWarning (
                    "Failed to locate gluErrorString() in %s", ogGLUPath
                );
        }
        triedToDLOpen = 1;
    }
#endif
    while ((error = glGetError ()) != GL_NO_ERROR)
        ogWarning ("GL error: %s", gluErrorString (error));
}

/*!
    \fn
    \brief    Set autorepeat status.
    \ingroup  inputstate
    \param    ignore    Whether to ignore autorepeated keys.

              If \a ignore is non-zero, then auto-repeat is
              disabled for keyboard callbacks for the
              <i>current window</i>.

              It is a fatal error to call this function without
              first initializing OpenGLEAN via glutInit(),
              and also it is fatal to call this function without
              having a <i>current window</i>.

    \see      glutSetKeyRepeat()
*/
void OGAPIENTRY glutIgnoreKeyRepeat (int ignore)
{
    OPENGLEAN_REQUIRE_READY ("glutIgnoreKeyRepeat");
    OPENGLEAN_REQUIRE_WINDOW ("glutIgnoreKeyRepeat");
    ogStructure.Window->State.IgnoreKeyRepeat = ignore ? GL_TRUE : GL_FALSE;
}

/*!
    \fn
    \brief    Sets autorepeat behavior for all OpenGLEAN windows.
    \ingroup  inputstate
    \param    repeatMode    On, Off or Default.

              glutSetKeyRepeat() is similar to glutIgnoreKeyRepeat()
              but sets the behavior
              for OpenGLEAN in general, rather than for a particular
              window.  The options for \a repeatMode are:

               - \a GLUT_KEY_REPEAT_OFF     \n Turn off repeat for all windows.
               - \a GLUT_KEY_REPEAT_ON      \n Turn on repeat for all windows.
               - \a GLUT_KEY_REPEAT_DEFAULT \n Respect the window's setting.

              It is a fatal error to call this function without first
              initializing OpenGLEAN via glutInit().  It is also
              fatal to call this function with a \a repeatMode value
              that is out of range.  The former could possibly be
              eliminated; the latter should probably be reduced to
              a warning or inform.

    \see      glutIgnoreKeyRepeat()
*/
void OGAPIENTRY glutSetKeyRepeat (int repeatMode)
{
    OPENGLEAN_REQUIRE_READY ("glutSetKeyRepeat");
    switch (repeatMode)
    {
    case GLUT_KEY_REPEAT_OFF:
    case GLUT_KEY_REPEAT_ON:
        ogState.KeyRepeat = repeatMode;
        break;

    case GLUT_KEY_REPEAT_DEFAULT:
        ogState.KeyRepeat = GLUT_KEY_REPEAT_ON;
        break;

    default:
        ogError ("Invalid glutSetKeyRepeat mode: %d", repeatMode);
        break;
    }
}

/*!
    \fn
    \brief    Forces a joystick poll and callback.
    \ingroup  deprecated

              THIS FUNCTION IS DEPRECATED.

              Like all OpenGLEAN joystick support, this
              function is deprecated.  The plan is to
              spin joystick support off into a separate
              library.

              Forces the OpenGLEAN joystick code to poll your
              joystick(s) and to call your joystick callbacks
              with the result.  The operation completes, including
              callbacks, before glutForceJoystickFunc() returns.

              It is a fatal error to call this function without
              first initializing OpenGLEAN via glutInit().

    \see      glutJoystickFunc()
*/
void OGAPIENTRY glutForceJoystickFunc (void)
{
    OPENGLEAN_REQUIRE_READY ("glutForceJoystickFunc");
    OPENGLEAN_DEPRECATED ("glutForceJoystickFunc()");
    if (ogStructure.Window && FETCH_WCB (*(ogStructure.Window), Joystick))
        ogJoystickPollWindow (ogStructure.Window);
}
