/*!
    \file  state.c
    \brief OpenGLEAN state query methods.
*/
/*
 * Copyright 2004, 2005, the OpenGLEAN Project.
 *
 * Portions copyright (C) 2004, the OpenGLUT project contributors.
 * OpenGLUT branched from freeglut in February, 2004.
 *
 * Copyright (c) 1999-2000 Pawel W. Olszta. All Rights Reserved.
 * Written by Pawel W. Olszta, <olszta@sourceforge.net>
 * Creation date: Thu Dec 16 1999
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * TODO BEFORE THE STABLE RELEASE:
 *
 *  glutGet()               -- X11 tests passed, but check if all enums
 *                             handled (what about Win32?)
 *  glutDeviceGet()         -- X11 tests passed, but check if all enums
 *                             handled (what about Win32?)
 *  glutGetModifiers()      -- OK, but could also remove the limitation
 *  glutLayerGet()          --
 */

/* -- PRIVATE FUNCTIONS ---------------------------------------------------- */

#if TARGET_HOST_UNIX_X11
/*
 * Queries the GL context about some attributes
 */
static int oghGetConfig (int attribute)
{
    int returnValue = 0;

    if (ogStructure.Window)
        glXGetConfig (
            ogDisplay.Display, ogStructure.Window->Window.VisualInfo,
            attribute, &returnValue
        );

    return returnValue;
}

#endif

/* -- INTERFACE FUNCTIONS -------------------------------------------------- */

/*!
    \fn
    \brief    Allows you to set some general state/option variables.
    \ingroup  state
    \param    eWhat    Enumerated parameter ID.
    \param    value    New value.

              Stores the \a value into a state variable named by
              \a eWhat.

              Allowable \a eWhat IDs are:

              - \a GLUT_ACTION_ON_WINDOW_CLOSE
                   (<b>DEPRECATED</b>)
                   A replacement method based on a new
                   window-close callback is expected.
                   Controls what happens when a window is closed by the
                   user or system.
                   \a GLUT_ACTION_EXIT
                   will immediately exit the application (default).
                   \a GLUT_ACTION_GLUTMAINLOOP_RETURNS
                   will immediately return from the main loop.
                   \a GLUT_ACTION_CONTINUE_EXECUTION
                   will contine execution of remaining windows.
              - \a GLUT_INIT_DISPLAY_MODE
                   An alternate way to set the display mode of
                   new windows.
              - \a GLUT_INIT_WINDOW_HEIGHT
                   An alternate way to set the height of new windows.
              - \a GLUT_INIT_WINDOW_WIDTH
                   An alternate way to set the width of new windows.
              - \a GLUT_INIT_WINDOW_X
                   An alternate way to set the initial horizontal position
                   of new windows.
              - \a GLUT_INIT_WINDOW_Y
                   An alternate way to set the initial vertical position
                   of new windows.
              - \a GLUT_RENDERING_CONTEXT
                   Set to either \a GLUT_CREATE_NEW_CONTEXT or
                   \a GUT_USE_CURRENT_CONTEXT to indicate
                   whether to share the current OpenGL rendering context
                   with new windows.
              - \a GLUT_WINDOW_CURSOR
                   Attempt to set the <i>current window</i>'s current cursor
                   as if by glutSetCursor().

              Additionally, the following are <b>experimental</b>.

              - \a GLUT_INIT_AUX_BUFFERS
                   Sets the number of aux. buffers for new windows.
                   (default is 0).

              It is an error to call this function without having first
              called glutInit().  Because this would be a pretty basic
              logic problem, and there is presently no other way to
              report errors, the error is fatal.

    \note     \a GLUT_INIT_AUX_BUFFERS on offscreen windows is undefined.
              (I can't think of a good reason to go either way on this
               issue.  On X11 systems with pbuffers, the code simply
               doesn't support aux. buffers for offscreen...  I should
               probably allow it and see what is required to make it
               work with pbuffers (probably trivial).)

    \see      glutGet(), glutDeviceGet(), glutGetModifiers(),
              glutLayerGet(), glutDestroyWindow(), glutMainLoop(),
              glutInitDisplayMode(), glutInit(), glutInitWindowSize(),
              glutInitWindowPosition(), glutSetCursor()
*/
void OGAPIENTRY glutSetOption (GLenum eWhat, int value)
{
    OPENGLEAN_REQUIRE_READY ("glutSetOption");
    switch (eWhat)
    {
    case GLUT_ACTION_ON_WINDOW_CLOSE:
        OPENGLEAN_DEPRECATED ("glutSetOption(GLUT_ACTION_ON_WINDOW)");
        ogState.ActionOnWindowClose = value;
        break;

    case GLUT_INIT_AUX_BUFFERS:
        ogState.AuxBuffers = value;
        break;

    case GLUT_INIT_DISPLAY_MODE:
        ogState.DisplayMode = (unsigned int) value;
        break;

    case GLUT_INIT_WINDOW_HEIGHT:
        ogState.Size.Y = (GLint) value;
        break;

    case GLUT_INIT_WINDOW_WIDTH:
        ogState.Size.X = (GLint) value;
        break;

    case GLUT_INIT_WINDOW_X:
        ogState.Position.X = (GLint) value;
        break;

    case GLUT_INIT_WINDOW_Y:
        ogState.Position.Y = (GLint) value;
        break;

    case GLUT_RENDERING_CONTEXT:
        if (GLUT_USE_CURRENT_CONTEXT == value)
            ogState.UseCurrentContext = GL_TRUE;
        else if (GLUT_CREATE_NEW_CONTEXT == value)
            ogState.UseCurrentContext = GL_FALSE;
        else
            ogWarning (
                "glutSetOption(): "
                "Unknown GLUT_RENDERING_CONTEXT parameter: %d",
                value
            );
        break;

    case GLUT_WINDOW_CURSOR:
        /* This check should be done by glutSetCursor()... */
        if (ogStructure.Window)
            glutSetCursor (value);
        break;

    default:
        ogWarning ("glutSetOption(): missing enum handle %i", eWhat);
        break;
    }
}

/*!
    \fn
    \brief    Allows you to query some general state/option variables.
    \ingroup  state
    \param    eWhat    Enumerated parameter ID.

              This function permits you to query for the current value
              of many different OpenGLEAN state variables.  The current
              list is:

              - \a GLUT_ACTION_ON_WINDOW_CLOSE
                   (<b>DREPRECATED</b>)
                   Allows you to do something other than die if the
                   user closes one of your windows.
              - \a GLUT_DISPLAY_MODE_POSSIBLE
                   Returns 0 if the currently requested window mode
                   is infeasible.  ("Display mode" is an unfortunate
                   term, as things such as auxiliary buffers, which are
                   not part of the display mode bit-flags, will affect
                   this.)  Success here does not guarantee that the window
                   can open, but failure suggests that you should try
                   with less ambitious feature requests.
              - \a GLUT_ELAPSED_TIME
                   Returns the milliseconds since OpenGLEAN initialized its
                   timer (generally glutInit() will do this, but requesting
                   the elapsed time will also effectively initialize the
                   timer if that was not already done).
              - \a GLUT_INIT_DISPLAY_MODE
                   Returns the bitmask corresponding to glutInitDisplayMode().
              - \a GLUT_INIT_STATE
                   Returns 0 if OpenGLEAN has not been initialized.
                   Otherwise returns non-zero.
              - \a GLUT_INIT_WINDOW_HEIGHT
                   Returns the current expected/requested height of new
                   windows.
              - \a GLUT_INIT_WINDOW_WIDTH
                   Returns the current expected/requested width of new windows.
              - \a GLUT_INIT_WINDOW_X
                   Returns the current expected/requested horizontal position
                   of new windows.
              - \a GLUT_INIT_WINDOW_Y
                   Returns the current expected/requested vertical position of
                   new windows.
              - \a GLUT_MENU_NUM_ITEMS
                   Returns the number of menu items on the <i>current menu</i>.
              - \a GLUT_RENDERING_CONTEXT
                   This variable controls context-sharing when OpenGLEAN
                   opens new windows.
              - \a GLUT_SCREEN_HEIGHT
                   Returns the height in pixels of the entire display; may
                   be larger than the current video mode.
              - \a GLUT_SCREEN_HEIGHT_MM
                   Returns the eight in millimeters of the entire display; may
                   be larger than the current video mode.
              - \a GLUT_SCREEN_WIDTH
                   Returns the width in pixels of the entire display; may
                   be larger than the current video mode.
              - \a GLUT_SCREEN_WIDTH_MM
                   Returns the width in millimeters of the entire display; may
                   be larger than the current video mode.
              - \a GLUT_VERSION
                   (<b>DEPRECATED</b>)
                   It's not clear whether this should return
                   a GLUT library version, or the OpenGLEAN
                   version.  Encoded as OPENGLEAN_VERSION, below.
              - \a OPENGLEAN_VERSION
                   The version of OpenGLEAN, encoded as
                   ( 10000*MAJOR ) + ( 100*MINOR ) + PATCH
              - \a GLUT_WINDOW_ACCUM_ALPHA_SIZE
                   Returns the number of bits to use for an alpha channel
                   in the accumulation buffer for the <i>current window</i>.
              - \a GLUT_WINDOW_ACCUM_BLUE_SIZE
                   Returns the number of bits to use for a blue channel
                   in the accumulation buffer for the <i>current window</i>.
              - \a GLUT_WINDOW_ACCUM_GREEN_SIZE
                   Returns the number of bits to use for a green channel
                   in the accumulation buffer for the <i>current window</i>.
              - \a GLUT_WINDOW_ACCUM_RED_SIZE
                   Returns the number of bits to use for a red channel
                   in the accumulation buffer for the <i>current window</i>.
              - \a GLUT_WINDOW_ALPHA_SIZE
                   Returns the number of bits to use for an alpha channel
                   for the color buffer in the <i>current window</i>.
              - \a GLUT_WINDOW_BLUE_SIZE
                   Returns the number of bits to use for a blue channel
                   for the color buffer in the <i>current window</i>.
              - \a GLUT_WINDOW_BORDER_WIDTH
                   Returns the pixel-width of the window border for
                   the <i>current window</i>
              - \a GLUT_WINDOW_BUFFER_SIZE
                   Returns the bit-depth of the <i>current window</i>.
              - \a GLUT_WINDOW_COLORMAP_SIZE
                   Currently returns 0 if you have an RGB window.
                   Otherwise, should return the number of colors in the
                   colormap.
              - \a GLUT_WINDOW_CURSOR
                   Returns the current GLUT cursor.
                   See glutSetCursor().
              - \a GLUT_WINDOW_DEPTH_SIZE
                   Returns the number of bits in the depth buffer for the
                   <i>current window</i>.
              - \a GLUT_WINDOW_DOUBLEBUFFER
                   Returns 0 if the current window is single-buffered;
                   else returns non-zero.
              - \a GLUT_WINDOW_FORMAT_ID
                   System dependant.
              - \a GLUT_WINDOW_GREEN_SIZE
                   Returns the number of bits to use for a green channel
                   for the color buffer in the <i>current window</i>.
              - \a GLUT_WINDOW_HEADER_HEIGHT
                   Returns the pixel-height of the titlebar on top of the
                   <i>current window</i>.
                   (It's not clear if this includes the border, nor what
                   an application should do to deal with title bars placed
                   on the sides of windows, as some windowmanagers do.)
              - \a GLUT_WINDOW_HEIGHT
                   Returns the pixel height of the <i>current window</i>.
              - \a GLUT_WINDOW_NUM_CHILDREN
                   Returns the number of children of the <i>current window</i>.
                   It's not clear if this should include menu windows.
              - \a GLUT_WINDOW_NUM_SAMPLES
                   Returns the number of samples used for multisampling for
                   the <i>current window</i>.
              - \a GLUT_WINDOW_PARENT
                   Returns the GLUT window id of the parent window of
                   the <i>current window</i>.  If there is
                   no <i>current window</i>, or if it has no parent, then
                   0 is returned.
              - \a GLUT_WINDOW_RED_SIZE
                   Returns the number of bits to use for a red channel
                   for the color buffer in the <i>current window</i>.
              - \a GLUT_WINDOW_RGBA
                   Returns 0 if the color buffer for the <i>current window</i>
                   stores indexed colors (a.k.a. paletted colors); returns
                   non-zero if the <i>current window</i> stores RGB information
                   in each pixel.
              - \a GLUT_WINDOW_STENCIL_SIZE
                   Returns the bit-depth of the stencil buffer for
                   the <i>current window</i>.
              - \a GLUT_WINDOW_STEREO
                   Returns non-zero if the color buffers for
                   the <i>current window</i> exist in left/right pairs.
                   Else returns 0.
              - \a GLUT_WINDOW_WIDTH
                   Returns the pixel width of the <i>current window</i>.
              - \a GLUT_WINDOW_X
                   Returns the horizontal pixel position of the
                   <i>current window</i>.
              - \a GLUT_WINDOW_Y
                   Returns the vertical pixel position of the
                   <i>current window</i>.

              Most of the above are very obvious, and so full documentation
              is postponed for now.

              Additionally, the following are <b>experimental</b>.

              - \a GLUT_INIT_AUX_BUFFERS
                   Returns the number of aux. buffers for new
                   windows.

              With the exception of \a GLUT_INIT_STATE
              and \a GLUT_ELAPSED_TIME, it is a fatal error to call this
              function without properly initializing OpenGLEAN via
              glutInit().  If the parameter to glutGet() is unknown,
              a return of -1 is given, but a -1 return value is not
              necessarily an error, as window positions may be negative.

    \note     \a GLUT_WINDOW_DEPTH_SIZE is reported one way on glX, using
              glXGetConfig(), and another way for WIN32
              (using glGetIntegerv()).  Probably we should unify these.

    \note     \a GLUT_WINDOW_DOUBLE_BUFFER can probably be unified
              to use glGetBoolean(), or even to check OpenGLEAN internal
              state.

    \see      glutSetOption(), glutDeviceGet(), glutGetModifiers(),
              glutLayerGet()

    \todo     Multisample support (\a GLUT_WINDOW_NUM_SAMPLES, here).

    \todo     \a GLUT_WINDOW_BUFFER_SIZE should be implemented for WIN32.

    \todo     Add \a GLUT_WINDOW_AUX_BUFFERS?

    \todo     It looks like we could use an \a OPENGLEAN_REQUIRE_WINDOW() macro
              invocation in addition to \a OPENGLEAN_REQUIRE_READY().

    \todo     \a GLUT_DISPLAY_MODE_POSSIBLE does not need us to run
              \a OPENGLEAN_REQUIRE_READY(), does it?

    \todo     This function is a bit messy, especially the WINCE part.  Fix.

    \todo     Lots of code uses return to hop out.  Since it's such a
              sprawling function, it's easy to be in the middle and not
              be 100% sure if there's anything important at the end of
              the function, or if it is safe to just "drop out" of
              the current case and head for the bottom.
*/
int OGAPIENTRY  glutGet (GLenum eWhat)
{
    int ret = -1;

#if TARGET_HOST_WIN32 || TARGET_HOST_WINCE
    GLboolean boolValue;

#endif

    switch (eWhat)
    {
    case GLUT_INIT_STATE:
        return ogState.Initialised;

    case GLUT_ELAPSED_TIME:
        /*!
          \bug  ogElapsedTime() returns a \a long, but glutGet() only is
          defined to return an \a int.
        */
        return ogElapsedTime ();
    }

    OPENGLEAN_REQUIRE_READY ("glutGet");

    /* XXX In no meaningful or useful order */
    switch (eWhat)
    {
    case GLUT_SCREEN_WIDTH:       return ogDisplay.ScreenWidth;
    case GLUT_SCREEN_HEIGHT:      return ogDisplay.ScreenHeight;
    case GLUT_SCREEN_WIDTH_MM:    return ogDisplay.ScreenWidthMM;
    case GLUT_SCREEN_HEIGHT_MM:   return ogDisplay.ScreenHeightMM;
    case GLUT_INIT_WINDOW_X:      return ogState.Position.X;
    case GLUT_INIT_WINDOW_Y:      return ogState.Position.Y;
    case GLUT_INIT_WINDOW_WIDTH:  return ogState.Size.X;
    case GLUT_INIT_WINDOW_HEIGHT: return ogState.Size.Y;
    case GLUT_INIT_AUX_BUFFERS:   return ogState.AuxBuffers;
    case GLUT_INIT_DISPLAY_MODE:  return ogState.DisplayMode;

        /*
         * The window/context specific queries are handled mostly by
         * oghGetConfig().
         */
    case GLUT_WINDOW_NUM_SAMPLES:
        return 0; /* XXX Return what I know about multisampling. */

#if TARGET_HOST_UNIX_X11
        /*
         * The rest of GLX queries under X are general enough to use a macro to
         * check them
         */
#define GLX_QUERY(a,b) case a: return oghGetConfig( b );

        GLX_QUERY (GLUT_WINDOW_RGBA, GLX_RGBA);
        GLX_QUERY (GLUT_WINDOW_DOUBLEBUFFER, GLX_DOUBLEBUFFER);
        GLX_QUERY (GLUT_WINDOW_BUFFER_SIZE, GLX_BUFFER_SIZE);
        GLX_QUERY (GLUT_WINDOW_STENCIL_SIZE, GLX_STENCIL_SIZE);
        GLX_QUERY (GLUT_WINDOW_DEPTH_SIZE, GLX_DEPTH_SIZE);
        GLX_QUERY (GLUT_WINDOW_RED_SIZE, GLX_RED_SIZE);
        GLX_QUERY (GLUT_WINDOW_GREEN_SIZE, GLX_GREEN_SIZE);
        GLX_QUERY (GLUT_WINDOW_BLUE_SIZE, GLX_BLUE_SIZE);
        GLX_QUERY (GLUT_WINDOW_ALPHA_SIZE, GLX_ALPHA_SIZE);
        GLX_QUERY (GLUT_WINDOW_ACCUM_RED_SIZE, GLX_ACCUM_RED_SIZE);
        GLX_QUERY (GLUT_WINDOW_ACCUM_GREEN_SIZE, GLX_ACCUM_GREEN_SIZE);
        GLX_QUERY (GLUT_WINDOW_ACCUM_BLUE_SIZE, GLX_ACCUM_BLUE_SIZE);
        GLX_QUERY (GLUT_WINDOW_ACCUM_ALPHA_SIZE, GLX_ACCUM_ALPHA_SIZE);
        GLX_QUERY (GLUT_WINDOW_STEREO, GLX_STEREO);

#undef GLX_QUERY

        /* Colormap size is handled in a bit different way than all the rest */
    case GLUT_WINDOW_COLORMAP_SIZE:
        if ((oghGetConfig (GLX_RGBA)) || (ogStructure.Window == NULL))
            /*
             * XXX Is this correct?  The X colormap support lets you *read*
             * XXX a "colormap" in this case, but it is a read-only map.
             *
             * We've got a RGBA visual, so there is no colormap at all.
             * The other possibility is that we have no current window set.
             */
            return 0;
        return ogStructure.Window->Window.VisualInfo->visual->map_entries;

        /*
         * Those calls are somewhat similiar, as they use XGetWindowAttributes()
         * function
         */
    case GLUT_WINDOW_X:
    case GLUT_WINDOW_Y:
    case GLUT_WINDOW_BORDER_WIDTH:
    case GLUT_WINDOW_HEADER_HEIGHT:
        {
            int x, y;
            Window w;

            if (ogStructure.Window == NULL)
                return 0;

            XTranslateCoordinates (
                ogDisplay.Display,
                ogStructure.Window->Window.Handle,
                ogDisplay.RootWindow,
                0, 0, &x, &y, &w
            );

            switch (eWhat)
            {
            case GLUT_WINDOW_X: return x;
            case GLUT_WINDOW_Y: return y;
            }

            if (w == 0)
                return 0;
            XTranslateCoordinates (
                ogDisplay.Display,
                ogStructure.Window->Window.Handle,
                w, 0, 0, &x, &y, &w
            );

            switch (eWhat)
            {
            case GLUT_WINDOW_BORDER_WIDTH:  return x;
            case GLUT_WINDOW_HEADER_HEIGHT: return y;
            }
            break;
        }

    case GLUT_WINDOW_WIDTH:
    case GLUT_WINDOW_HEIGHT:
        {
            XWindowAttributes winAttributes;

            if (ogStructure.Window->State.IsOffscreen)
                switch (eWhat)
                {
                case GLUT_WINDOW_WIDTH:
                    return ogStructure.Window->State.Width;
                case GLUT_WINDOW_HEIGHT:
                    return ogStructure.Window->State.Height;
                }

            if (ogStructure.Window == NULL)
                return 0;
            XGetWindowAttributes (
                ogDisplay.Display,
                ogStructure.Window->Window.Handle,
                &winAttributes
            );
            switch (eWhat)
            {
            case GLUT_WINDOW_WIDTH:  return winAttributes.width;
            case GLUT_WINDOW_HEIGHT: return winAttributes.height;
            }
            break;
        }

        /* Not sure if there will be a ogChooseVisual() function for Win32 */
    case GLUT_DISPLAY_MODE_POSSIBLE: return !!ogChooseVisual ();

        /* This is system-dependant */
    case GLUT_WINDOW_FORMAT_ID:
        if (!ogStructure.Window)
            return 0;
        /*
           ! \bug The following returns an unsigned long, but the function is
           only set to return {int}.
        */
        return ogStructure.Window->Window.VisualInfo->visualid;

#elif TARGET_HOST_WIN32 || TARGET_HOST_WINCE

        /* Handle the OpenGL inquiries */
    case GLUT_WINDOW_RGBA:
        glGetBooleanv (GL_RGBA_MODE, &boolValue);
        ret = boolValue ? 1 : 0;
        return ret;
    case GLUT_WINDOW_DOUBLEBUFFER:
        glGetBooleanv (GL_DOUBLEBUFFER, &boolValue);
        ret = boolValue ? 1 : 0;
        return ret;
    case GLUT_WINDOW_STEREO:
        glGetBooleanv (GL_STEREO, &boolValue);
        ret = boolValue ? 1 : 0;
        return ret;

    case GLUT_WINDOW_RED_SIZE:
        glGetIntegerv (GL_RED_BITS, &ret);
        return ret;
    case GLUT_WINDOW_GREEN_SIZE:
        glGetIntegerv (GL_GREEN_BITS, &ret);
        return ret;
    case GLUT_WINDOW_BLUE_SIZE:
        glGetIntegerv (GL_BLUE_BITS, &ret);
        return ret;
    case GLUT_WINDOW_ALPHA_SIZE:
        glGetIntegerv (GL_ALPHA_BITS, &ret);
        return ret;
    case GLUT_WINDOW_ACCUM_RED_SIZE:
        glGetIntegerv (GL_ACCUM_RED_BITS, &ret);
        return ret;
    case GLUT_WINDOW_ACCUM_GREEN_SIZE:
        glGetIntegerv (GL_ACCUM_GREEN_BITS, &ret);
        return ret;
    case GLUT_WINDOW_ACCUM_BLUE_SIZE:
        glGetIntegerv (GL_ACCUM_BLUE_BITS, &ret);
        return ret;
    case GLUT_WINDOW_ACCUM_ALPHA_SIZE:
        glGetIntegerv (GL_ACCUM_ALPHA_BITS, &ret);
        return ret;
    case GLUT_WINDOW_DEPTH_SIZE:
        glGetIntegerv (GL_DEPTH_BITS, &ret);
        return ret;

    case GLUT_WINDOW_BUFFER_SIZE:
        ret = 1;    /* ????? */
        return ret;
    case GLUT_WINDOW_STENCIL_SIZE:
        ret = 0;    /* ????? */
        return ret;

    case GLUT_WINDOW_X:
    case GLUT_WINDOW_Y:
    case GLUT_WINDOW_WIDTH:
    case GLUT_WINDOW_HEIGHT:
        {
            RECT winRect;

            ret = 0;
            if (ogStructure.Window)
            {
                if (ogStructure.Window->State.IsOffscreen)
                    switch (eWhat)
                    {
                    case GLUT_WINDOW_WIDTH:
                        return ogStructure.Window->State.Width;
                    case GLUT_WINDOW_HEIGHT:
                        return ogStructure.Window->State.Height;
                    }
                /*
                 * We need to call GetWindowRect() first...
                 *  (this returns the pixel coordinates of the outside of
                 *   the window)
                 */
                GetWindowRect (ogStructure.Window->Window.Handle, &winRect);

                /* we must correct the results we've just received */
#if !TARGET_HOST_WINCE
                if ((ogStructure.GameMode != ogStructure.Window) &&
                        (ogStructure.Window->Parent == NULL) &&
                        (!ogStructure.Window->IsUnmanaged))
                {
                    winRect.left += GetSystemMetrics (SM_CXSIZEFRAME);
                    winRect.right -= GetSystemMetrics (SM_CXSIZEFRAME);
                    winRect.top += GetSystemMetrics (SM_CYSIZEFRAME) +
                        GetSystemMetrics (SM_CYCAPTION);
                    winRect.bottom -= GetSystemMetrics (SM_CYSIZEFRAME);
                }
#endif

                switch (eWhat)
                {
                case GLUT_WINDOW_X:
                    return winRect.left;
                case GLUT_WINDOW_Y:
                    return winRect.top;
                case GLUT_WINDOW_WIDTH:
                    return winRect.right - winRect.left;
                case GLUT_WINDOW_HEIGHT:
                    return winRect.bottom - winRect.top;
                }
            }
        }
        break;

    case GLUT_WINDOW_BORDER_WIDTH:
#if TARGET_HOST_WINCE
        return 0;
#else
        return GetSystemMetrics (SM_CXSIZEFRAME);
#endif

    case GLUT_WINDOW_HEADER_HEIGHT:
#if TARGET_HOST_WINCE
        return 0;
#else
        return GetSystemMetrics (SM_CYCAPTION);
#endif

    case GLUT_DISPLAY_MODE_POSSIBLE:
#if TARGET_HOST_WINCE
        return GL_FALSE;
#else
        return ogSetupPixelFormat (
            ogStructure.Window, GL_TRUE,
            PFD_MAIN_PLANE
        );
#endif

    case GLUT_WINDOW_FORMAT_ID:
#if !TARGET_HOST_WINCE
        if (ogStructure.Window != NULL)
            return GetPixelFormat (ogStructure.Window->Window.Device);
#endif
        return 0;

#endif

        /*
         * The window structure queries
         */
    case GLUT_WINDOW_PARENT:
        if (ogStructure.Window == NULL)
            return 0;
        if (ogStructure.Window->Parent == NULL)
            return 0;
        return ogStructure.Window->Parent->ID;

    case GLUT_WINDOW_NUM_CHILDREN:
        if (ogStructure.Window == NULL)
            return 0;
        return ogListLength (&ogStructure.Window->Children);

    case GLUT_WINDOW_CURSOR:
        if (ogStructure.Window == NULL)
            return 0;
        return ogStructure.Window->State.Cursor;

    case GLUT_MENU_NUM_ITEMS:
        if (ogStructure.Menu == NULL)
            return 0;
        return ogListLength (&ogStructure.Menu->Entries);

    case GLUT_ACTION_ON_WINDOW_CLOSE:
        OPENGLEAN_DEPRECATED ("glutGet(GLUT_ACTION_ON_WINDOW_CLOSE)");
        return ogState.ActionOnWindowClose;

    case GLUT_VERSION:
        OPENGLEAN_DEPRECATED ("glutGet(GLUT_VERSION)");
    case OPENGLEAN_VERSION:
        return VERSION_MAJOR * 10000 + VERSION_MINOR * 100 + VERSION_PATCH;

    case GLUT_RENDERING_CONTEXT:
        return ogState.UseCurrentContext ?
            GLUT_USE_CURRENT_CONTEXT :
            GLUT_CREATE_NEW_CONTEXT;

    default:
        ogWarning ("glutGet(): missing enum handler %d", eWhat);
        break;
    }
    return ret;
}

/*!
    \fn
    \brief    Allows you to get some device state/option variables.
    \ingroup  inputstate
    \param    eWhat    Enumerated parameter ID.

              Retrieves some system-specific information about
              attached devices.  Supported device queries are:

              - \a GLUT_HAS_KEYBOARD
                   Return non-zero if there is a keyboard.
              - \a GLUT_HAS_MOUSE
                   Return non-zero if there is a mouse.
              - \a GLUT_HAS_SPACEBALL
                   (<b>DEPRECATED</b>)
                   Return non-zero if there is a spaceball.
              - \a GLUT_HAS_JOYSTICK
                   (<b>DEPRECATED</b>)
                   Checks if a joystick is available.
              - \a GLUT_OWNS_JOYSTICK
                   (<b>DEPRECATED</b>)
                   Checks if OpenGLEAN can use it.
              - \a GLUT_JOYSTICK_AXES
                   (<b>DEPRECATED</b>)
                   Return the number of axes for the joystick.
              - \a GLUT_JOYSTICK_POLL_RATE
                   (<b>DEPRECATED</b>)
                   Return the rate (in GLUT timer ticks?) at
                   which the joystick is polled.  Requires a
                   current window to be set.
              - \a GLUT_NUM_MOUSE_BUTTONS
                   Return the number of buttons that the user's
                   mouse has.
              - \a GLUT_OWNS_JOYSTICK
                   (<b>DEPRECATED</b>)
                   Return non-zero if OpenGLEAN believes that it has
                   successfully acquired access to the joystick.
              - \a GLUT_DEVICE_IGNORE_KEY_REPEAT
                   Return non-zero if the <i>current window</i> is
                   set to disable key repeating.
                   (Contrast \a GLUT_DEVICE_KEY_REPEAT.)
              - \a GLUT_DEVICE_KEY_REPEAT
                   Returns the key repeat mode, one of:
                   \a GLUT_KEY_REPEAT_OFF,
                   \a GLUT_KEY_REPEAT_ON, or
                   \a GLUT_KEY_REPEAT_DEFAULT.
                   This is apparently independant of your current
                   window's setting.
                   (Contrast \a GLUT_DEVICE_IGNORE_KEY_REPEAT.)
              - \a GLUT_HAS_DIAL_AND_BUTTON_BOX
                   (<b>DEPRECATED</b>)
                   Return non-zero if a dials-and-buttons box is
                   present.
              - \a GLUT_HAS_TABLET
                   (<b>DEPRECATED</b>)
                   Return non-zero if a tablet is present.
              - \a GLUT_NUM_BUTTON_BOX_BUTTONS
                   (<b>DEPRECATED</b>)
                   Return the number of buttons on a dials-and-buttons
                   box, if any.
              - \a GLUT_NUM_DIALS
                   (<b>DEPRECATED</b>)
                   Return the number of dials on a dials-and-buttons
                   box, if any.
              - \a GLUT_NUM_SPACEBALL_BUTTONS
                   (<b>DEPRECATED</b>)
                   Return the number of buttons on a spaceball, if any.
              - \a GLUT_NUM_TABLET_BUTTONS
                   (<b>DEPRECATED</b>)
                   Return the number of buttons on a tablet, if any.

              It is an error---presently fatal---to call this function
              without first properly initializing OpenGLEAN via
              glutInit().  If the parameter to glutDeviceGet() is
              unknown, a return value of -1 is given.  It is believed
              that at present a -1 return value always indicates an
              error, but that may not be a safe assumption forever.

    \note     freeglut and OpenGLUT confusingly claim that the
              \a GLUT_DEVICE_KEY_REPEAT get will return the
              repeat rate.  This is wrong, both with respect
              to their own implementation and with respect to old
              GLUT.

    \see      glutSetOption(), glutGet(), glutGetModifiers(),
              glutLayerGet()

    \todo     Consider moving to a table-based approach rather than a
              switch(), letting us move to modular functions.

    \bug      Keyboards are optional, but OpenGLEAN doesn't detect
              their absence.

    \bug      Mice are optional, but OpenGLEAN is only able to check
              for them under WIN32.

    \bug      Mice can have a varying number of buttons, but OpenGLEAN
              assumes exactly 3 on UNIX_X11.

    \bug      \a HAS/\a OWNS queries on joysticks don't report how many
              joysticks are there.

    \bug      \a BUTTONS/\a AXES queres on joysticks only work with
              the first joystick found.

    \bug      Only supports querying for one joystick.
*/
int OGAPIENTRY  glutDeviceGet (GLenum eWhat)
{
    int ret = -1;

    OPENGLEAN_REQUIRE_READY ("glutDeviceGet");
    switch (eWhat)
    {
    case GLUT_HAS_KEYBOARD:
        ret = TRUE;
        break;

#if TARGET_HOST_UNIX_X11

    case GLUT_HAS_MOUSE:
        ret = TRUE;
        break;

    case GLUT_NUM_MOUSE_BUTTONS:
        /*
         * Return the number of mouse buttons available. This is a big guess.
         *
         * The present heuristic is to assume that all mice ever
         * produced (future, present, and past) have exactly
         * 3 buttons.  This is obviously wrong.
         * There is no better alternative proposed at this time.
         * (But since this part is X-specific, we might use an env-var,
         *  command-line param, or even X-server- or OS-specific inquiry.
         */
        ret = 3;
        break;

#elif TARGET_HOST_WIN32 || TARGET_HOST_WINCE

    case GLUT_HAS_MOUSE:
        ret = GetSystemMetrics (SM_MOUSEPRESENT);
        break;

    case GLUT_NUM_MOUSE_BUTTONS:
#if TARGET_HOST_WINCE
        ret = 1;
#else
        ret = GetSystemMetrics (SM_CMOUSEBUTTONS);
#endif
        break;

#endif

#define GDG_WARN(sym) \
    OPENGLEAN_DEPRECATED ("glutDeviceGet(" #sym ")");
    case GLUT_HAS_JOYSTICK:
        GDG_WARN (GLUT_HAS_JOYSTICK);
        ret = ogJoystickDetect ();
        break;
    case GLUT_OWNS_JOYSTICK:
        GDG_WARN (GLUT_OWNS_JOYSTICK);
        ret = ogState.JoysticksInitted;
        break;
    case GLUT_JOYSTICK_POLL_RATE:
        GDG_WARN (GLUT_JOYSTICK_POLL_RATE);
        OPENGLEAN_REQUIRE_WINDOW ("glutDeviceGet");
        ret = ogStructure.Window->State.JoystickPollRate;
        break;
    case GLUT_JOYSTICK_BUTTONS:
        GDG_WARN (GLUT_JOYSTICK_BUTTONS);
        ret = glutGetJoystickNumButtons (0);
        break;
    case GLUT_JOYSTICK_AXES:
        GDG_WARN (GLUT_JOYSTICK_AXES);
        ret = glutGetJoystickNumAxes (0);
        break;

    case GLUT_HAS_SPACEBALL:
        GDG_WARN ("GLUT_HAS_SPACEBALL");
        ret = FALSE;
    case GLUT_HAS_DIAL_AND_BUTTON_BOX:
        GDG_WARN ("GLUT_HAS_DIAL_AND_BUTTON_BOX");
        ret = FALSE;
    case GLUT_HAS_TABLET:
        GDG_WARN ("GLUT_HAS_TABLET");
        ret = FALSE;
        break;

    case GLUT_NUM_SPACEBALL_BUTTONS:
        GDG_WARN ("GLUT_NUM_SPACEBALL_BUTTONS");
        ret = 0;
    case GLUT_NUM_BUTTON_BOX_BUTTONS:
        GDG_WARN ("GLUT_NUM_SPACEBALL_BUTTONS");
        ret = 0;
    case GLUT_NUM_DIALS:
        GDG_WARN ("GLUT_NUM_SPACEBALL_BUTTONS");
        ret = 0;
    case GLUT_NUM_TABLET_BUTTONS:
        GDG_WARN ("GLUT_NUM_SPACEBALL_BUTTONS");
        ret = 0;
        break;
#undef GDG_WARN

    case GLUT_DEVICE_IGNORE_KEY_REPEAT:
        ret = ogStructure.Window ?
            ogStructure.Window->State.IgnoreKeyRepeat :
            0;
        break;

    case GLUT_DEVICE_KEY_REPEAT:
        ret = ogState.KeyRepeat;
        break;

    default:
        ogWarning ("glutDeviceGet(): Unknown request type %d", eWhat);
        break;
    }

    return ret;
}

/*!
    \fn
    \brief    Returns the status of Alt, Shift, and Ctrl keys.
    \ingroup  inputstate

              According to which, if any, modifier keys are held,
              the return value is the logical \a OR combination
              of any of the following symbolic bitmasks:

               - \a GLUT_ACTIVE_SHIFT
               - \a GLUT_ACTIVE_CTRL
               - \a GLUT_ACTIVE_ALT

              E.g., if the shift key is held, and no other modifier
              keys are held, this function will return \a GLUT_ACTIVE_SHIFT.

              It is a fatal error to call this routine without properly
              initializing OpenGLEAN via glutInit().  It should
              only be called while servicing an OpenGLEAN input
              callback (e.g., mouse or key event); if you violate
              this, a warning will be printed and a 0 mask will be
              returned.  It is probably acceptable to drop the warning.
              The rationale
              for only calling withint an input callback is not
              clear.

    \note     Complains if not invoked by a client callback.

    \note     Old GLUT 3.7 describes this as glutCurrentModifiers(),
              even in the __glutWarning() message if you call it
              outside of an input callback.  However, even GLUT actually
              *defined* this function glutGetModifiers().

    \note     This function probably should be replaced with an enumerated
              constant passed into glutGet().  This would be more
              consistant with the OpenGL approach, and with, I think, most
              GLUT-like state retrieval.

    \note     Even though there is nothing dangerous about this function
              if we are not initialized, we do an error-exit if we are
              not "ready".  It is not terribly meaningful to call
              this function in that case.

    \see      glutSetOption(), glutGet(), glutDeviceGet(),
              glutLayerGet()

    \bug      Does not differentiate between the left and right forms
              of the modifiers.
*/
int OGAPIENTRY  glutGetModifiers (void)
{
    int ret = 0;

    OPENGLEAN_REQUIRE_READY ("glutGetModifiers");
    if (OG_STATE_INVALID_MODIFIERS == ogState.Modifiers)
        ogWarning ("glutGetModifiers() called outside an input callback");
    else
        ret = ogState.Modifiers;

    return ret;
}

/*!
    \fn
    \brief    Allows you to get some overlay state/option variables.
    \ingroup  overlays
    \param    eWhat    Enumerated parameter ID.

              Returns some useful information about layers.
              Or, it would be useful if layers were implemented...
              \a eWhat may be given any of the following values:

              - \a GLUT_HAS_OVERLAY
              - \a GLUT_LAYER_IN_USE
              - \a GLUT_NORMAL_DAMAGED
                 0 unless the window system has told us that
                 the normal layer is damaged (glutPostRedisplay()
                 does not affect this).
              - \a GLUT_OVERLAY_DAMAGED
                Returns -1 if no layer in use.
              - \a GLUT_OVERLAY_POSSIBLE
              - \a GLUT_TRANSPARENT_INDEX
                Returns -1 if no layer in use.

              All information relates to the <i>current window</i>
              and any overlay that it may have.

              It is a fatal error to call this function without
              properly initializing OpenGLEAN via glutInit().
              A -1 is returned if the parameter to glutLayerGet() is
              unknown; in certain other cases noted above, a -1 is
              also returned if no layer is in use.  The fatal error
              can probably be dropped, though the code is a little
              sloppy about \a FALSE, -1, and valid return values.
              Here, -1 seems to be the preferred way to indicate
              \a TRUE, though it is also the default value.

    \see      glutSetOption(), glutGet(), glutDeviceGet(), glutGetModifiers()
*/
int OGAPIENTRY  glutLayerGet (GLenum eWhat)
{
    int ret = -1;

    OPENGLEAN_REQUIRE_READY ("glutLayerGet");

    /*!
      \todo  Probably we should merge the below sections unless and until
      layer support is actually added.
    */
    switch (eWhat)
    {

#if TARGET_HOST_UNIX_X11

    case GLUT_OVERLAY_POSSIBLE:
        ret = FALSE;
        break;

    case GLUT_LAYER_IN_USE:
        ret = GLUT_NORMAL;
        break;

    case GLUT_HAS_OVERLAY:
        ret = FALSE;
        break;

    case GLUT_TRANSPARENT_INDEX:
    case GLUT_OVERLAY_DAMAGED:
        break;

#elif TARGET_HOST_WIN32 || TARGET_HOST_WINCE

    case GLUT_OVERLAY_POSSIBLE:
        /*
          return ogSetupPixelFormat(
              ogStructure.Window, GL_TRUE,
              PFD_OVERLAY_PLANE
          );
        */
        ret = FALSE;
        break;

    case GLUT_LAYER_IN_USE:
        ret = GLUT_NORMAL;
        break;

    case GLUT_HAS_OVERLAY:
        ret = FALSE;
        break;

    case GLUT_TRANSPARENT_INDEX:
    case GLUT_OVERLAY_DAMAGED:
        break;

#endif

    case GLUT_NORMAL_DAMAGED:
        /*!
          \todo Should be set when the window is damaged; should not be set
          just for glutPostRedisplay().
        */
        ret = FALSE;
        break;

    default:
        ogWarning ("glutLayerGet(): missing enum handle %i", eWhat);
        break;
    }

    return ret;
}
