/*!
    \file  videoresize.c
    \brief Video resizing
*/

/*
 * Copyright 2004, 2005, the OpenGLEAN Project.
 *
 * Portions copyright 2004, the OpenGLUT project contributors.
 * OpenGLUT branched from freeglut in February, 2004.
 *
 * Copyright (c) 1999-2000 Pawel W. Olszta. All Rights Reserved.
 * Written by Pawel W. Olszta, <olszta@sourceforge.net>
 * Creation date: Thu Dec 16 1999
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/* -- INTERFACE FUNCTIONS -------------------------------------------------- */

/*!
    \fn
    \brief    Query OpenGLEAN about video resizing.
    \ingroup  videomode
    \param    eWhat    An enumerated tag.

              This function is only implemented in skeletal form.

              From GLUT 3.7's implementation, the following is a
              list of symbols that GLUT 3.7 supported.  These are
              not necessarily defined in OpenGLEAN at this time,
              nor is support presently available for them:

               - \a GLUT_VIDEO_RESIZE_POSSIBLE
                Old GLUT returns something called \a canVideoResize.
                This appears to be a run-time constant.
               - \a GLUT_VIDEO_RESIZE_IN_USE
                Old GLUT used this to determine if you had called
                glutSetupVideoResizing(), it seems, but had not since then
                called glutStopVideoResizing().
               - \a GLUT_VIDEO_RESIZE_X_DELTA
                Old GLUT appears to return \a init_x - \a curr_x.
                Or possibly the other way around...
               - \a GLUT_VIDEO_RESIZE_Y_DELTA
                Old GLUT appears to return \a init_y - \a curr_y.
                Or possibly the other way around...
               - \a GLUT_VIDEO_RESIZE_WIDTH_DELTA
                Old GLUT appears to return \a init_width - \a curr_width.
                Or possibly the other way around...
               - \a GLUT_VIDEORESIZE_HEIGHT_DELTA
                Old GLUT appears to return \a init_height - \a curr_height.
                Or possibly the other way around...
               - \a GLUT_VIDEO_REISZE_X
                Old GLUT appears to return the current \a x.
               - \a GLUT_VIDEO_RESIZE_Y
                Old GLUT appears to return the current \a y.
               - \a GLUT_VIDEO_RESIZE_WIDTH
                Old GLUT appears to return the current \a width.
               - \a GLUT_VIDEO_RESIZE_HEIGHT
                Old GLUT appears to return the current \a height.

              A -1 should be returned for unimplemented features, and
              a warning printed.

              Additionally, the \a GLUT_VIDEO_RESIZE_CHANNEL environment
              variable is consulted (value treated as a string representation
              of an integer).

    \see      glutVideoResizeGet(), glutSetupVideoResizing(),
              glutStopVideoResizing(), glutVideoResize(), glutVideoPan()
*/
int OGAPIENTRY glutVideoResizeGet (GLenum eWhat)
{
    ogWarning ("glutVideoResizeGet (%d): Unimplemented.", eWhat);
    return -1;
}

/*!
    \fn
    \brief    Initialized GLUT's video resizing support.
    \ingroup  videomode

              The apparent purpose of this function in old GLUT was twofold:
              First, to help old GLUT set up any minimal structures prior
              do doing actual resizes.  Second, so that if you were not
              going to be able to resize the video display, you would
              be able to find out early.  In the GLUT paradigm, of course,
              every operation succeeds or dies trying.

              Except for the "die (early) trying" paradigm of GLUT, I do
              not think that there is any reason that we need to use
              this function.  The actual resizing code can take care of
              initialization, on-demand.  If, like window resizing, etc.,
              we regard video resizing as a <b>request</b> to the
              window system (which may or may not be honored), there
              is no need at all for this routine, that I can see.
              (I suppose that the counter-balancing glutStopVideoResizing()
              provides a simpler way to restore the original video status,
              however...)

              This function may or may not be retained.

              Except with SGI extensions to glX, old GLUT would always
              produce a fatal error.
              (Old GLUT required: \a GLX_SGIX_video_resize)

              In all cases, if testing GLUT_VIDEO_RESIZE_POSSIBLE
              (see glutVideoResizeGet()) returns 0, a fatal error
              will occur.

              OpenGLEAN currently prints a warning if you call this
              function.

    \note     I'm not sure that this API entry point should be reatined.
    \see      glutVideoResizeGet(), glutSetupVideoResizing(),
              glutStopVideoResizing(), glutVideoResize(), glutVideoPan()
*/
void OGAPIENTRY glutSetupVideoResizing (void)
{
    ogWarning ("glutSetupVideoResizing (): Video resizing not possible.");
}

/*!
    \fn
    \brief    Restore the video size, cleanup OpenGLEAN resizing.
    \ingroup  videomode

              In old GLUT, would set GLUT_VIDEO_RESIZE_IN_USE to 0
              (see glutGetVideoResize()).

              This function should restore the video mode to whatever
              it was when OpenGLEAN began video resizing, if OpenGLEAN
              has been doing any resizing.  It should clean up the internal
              state in OpenGLEAN, if any, so that subsequent resizing can
              be done.

    \note     I'm not sure that this API entry point should be retained.
    \note     If multiple programs are fiddling with your video mode, it
              is not clear that we can reliably "restore" things to what
              they "should" be.
    \see      glutVideoResizeGet(), glutSetupVideoResizing(),
              glutStopVideoResizing(), glutVideoResize(), glutVideoPan()
*/
void OGAPIENTRY glutStopVideoResizing (void)
{
    ogWarning ("glutStopVideoResizing (): Unimplemented.");
}

/*!
    \fn
    \brief    Set the video size and panning.
    \ingroup  videomode
    \param    x    A horizontal position.
    \param    y    A vertical position.
    \param    w    A width.
    \param    h    A height.

              This function is unimplemeneted.

              This function appears to allow setting the video
              display rectangle to a sub-rectangle of the
              hardware buffer.  \a x and \a y are apparently the
              upper left corner, and \a w and \a h are the width
              and height of the rectangle.

              It is not clear what should be done if the values
              are invalid (negative \a x, for example).

    \note     There appear to be no safeguards in old GLUT against
              multiple programs requesting various video resizes.
    \note     Appears to be redundant of glutVideoPan()
    \see      glutVideoResizeGet(), glutSetupVideoResizing(),
              glutStopVideoResizing(), glutVideoResize(), glutVideoPan()
*/
void OGAPIENTRY glutVideoResize (int x, int y, int w, int h)
{
    ogWarning(
        "glutVideoResize (%d, %d, %d, %d): Unimplemented.", x, y, w, h
    );
}

/*!
    \fn
    \brief    Set the video size and panning.
    \ingroup  videomode
    \param    x    A horizontal position.
    \param    y    A vertical position.
    \param    w    A width.
    \param    h    A height.

              This function is unimplemeneted.

              This function appears to allow setting the video
              display rectangle to a sub-rectangle of the
              hardware buffer.  \a x and \a y are apparently the
              upper left corner, and \a w and \a h are the width
              and height of the rectangle.

              It is not clear what should be done if the values
              are invalid (negative \a x, for example).

    \note     There appear to be no safeguards in old GLUT against
              multiple programs requesting various video resizes.
    \note     Appears to be redundant of glutVideoResize().
    \see      glutVideoResizeGet(), glutSetupVideoResizing(),
              glutStopVideoResizing(), glutVideoResize(), glutVideoPan()
*/
void OGAPIENTRY glutVideoPan (int x, int y, int w, int h)
{
    ogWarning( "glutVideoPan (%d, %d, %d, %d): Unimplemented.", x, y, w, h);
}
