#include <GL/openglean.h>
#include "../tools/opengleanDaveWindows.h"

// DAVE
#include "Dave.h"
#include "DaveConfig.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <math.h>

#include "../tools/DiffTime.h" // for accurate time measurements (milliseconds)
#include "../tools/msSleep.h" // for unified sleep (milliseconds)

struct MouseStruct {
	int x, y;
	int oldX, oldY;
	bool buttonPressed[3];
} mouse = {0, 0, 0, 0, false, false, false};

int modifiers = 0;

struct CameraStruct {
	float x, y, z;
	float camAngle1, camAngle2;
} camera = {0, 6, 0, 0, 0};

DiffTime diffTime, totalTime;

struct WinUserStruct {
	// data that's different for each window goes here, i.e. texture IDs etc.
};

// forward declarations

void initialize();
void drawScene(float totalTime, float diffTime, WinUserStruct* wus);
void reshape(GLsizei width, GLsizei height);

// DAVE
DAVE::Dave dave;
struct UserSyncDataStruct {
	CameraStruct cam;
};


///////////////////////////////////////////////////////////////////////////////

void display() {
	// for synchronized drawing, draw to all buffers at callback of first window
	// that's why only a display callback for the first window is registered
	if (glutGetWindow() != dave.windowManager->windows.front().systemWindowId) return;

	float dts = 0;
	dts = diffTime.getSeconds();
	diffTime.start();

	for (std::vector<DAVE::DaveWindowManager::DBufferStruct>::iterator db=dave.windowManager->dBuffers.begin(); db!=dave.windowManager->dBuffers.end(); ++db) {
		DAVE::DaveWindowManager::WindowStruct* win = &dave.windowManager->windows[db->windowIndex];
		WinUserStruct* wus = (WinUserStruct*) win->winUserDataPtr;

		glutSetWindow(win->systemWindowId);
		if (glutGetWindow() != win->systemWindowId) {
			std::cout << "warning: window could not be set" << std::endl;
			continue;
		}
		if (db->quadDrawBuffer) glDrawBuffer(db->quadDrawBuffer);

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		// set projection matrix
		dave.setProjectionMatrix(&*db);

		glLoadIdentity();
		glTranslatef(camera.x, camera.y, camera.z);
		glColor3f(1,1,1);

		drawScene(totalTime.getSeconds(), dts, wus);

		// software edge blending mask
		dave.applyBlendMask(win->blendMaskImg);
	}

	// swap all buffers at the same time (almost) rather than directly after drawing
	for (std::vector<DAVE::DaveWindowManager::DBufferStruct>::iterator db=dave.windowManager->dBuffers.begin(); db!=dave.windowManager->dBuffers.end(); ++db) {
		DAVE::DaveWindowManager::WindowStruct* win = &dave.windowManager->windows[db->windowIndex];

		glutSetWindow(win->systemWindowId);
		glutSwapBuffers();
	}
}

void update() {
	UserSyncDataStruct* usd = (UserSyncDataStruct*) dave.getSyncBufferUserDataPtr();
	if (dave.isServer()) {
		memcpy(&usd->cam, &camera, sizeof(CameraStruct));
	}

	dave.update();

	if (dave.isClient()) {
		memcpy(&camera, &usd->cam, sizeof(CameraStruct));
	}

	msSleep(2);

	//glutPostRedisplay();
	glutPostWindowRedisplay(dave.windowManager->windows.front().systemWindowId);
}

void init() {
	// print usage information
	printf("drag mouse to move or rotate camera\n");
	printf(" - rotate: left button\n");
	printf(" - move in x/y plane: middle button / left button + CTRL key\n");
	printf(" - move in z/zoom: right button / left button + SHIFT key / mouse wheel for zoom only\n");
	printf("press [esc] to exit\n\n");

	// initializations for each window

	for (std::vector<DAVE::DaveWindowManager::WindowStruct>::iterator win=dave.windowManager->windows.begin(); win!=dave.windowManager->windows.end(); ++win) {
		glutSetWindow(win->systemWindowId);

		glEnable(GL_DEPTH_TEST);
		glClearColor(0.0, 0.0, 0.0, 0.0);

		WinUserStruct* wus = new WinUserStruct;
		win->winUserDataPtr = wus;
	}

	diffTime.start();
	totalTime.start();

	update();
}

// letters and some other characters
void keyboardFunc(unsigned char key, int x, int y) {
	switch (key) {
		case 27: // escape
			exit(0);
			break;
   }
}

// mouse motion without button pressed
void mouseMove(int x, int y) {
	mouse.oldX = x;
	mouse.oldY = y;
}

// mouse motion with button pressed
void mouseDrag(int x, int y) {
	modifiers = glutGetModifiers();
	mouse.x = x;
	mouse.y = y;

	int dx = (mouse.x - mouse.oldX);
	int dy = (mouse.y - mouse.oldY);

	// rotate
	if (mouse.buttonPressed[GLUT_LEFT_BUTTON] && (modifiers == 0) ) {
		camera.camAngle1 += 0.4 * dy;
		camera.camAngle2 += 0.4 * dx;
	}

	// pan
	if ((mouse.buttonPressed[GLUT_LEFT_BUTTON] && (modifiers == GLUT_ACTIVE_CTRL))
		|| (mouse.buttonPressed[GLUT_MIDDLE_BUTTON])) {
		camera.x += 0.02 * dx;
		camera.z -= 0.02 * dy;
	}

	// zoom
	if ((mouse.buttonPressed[GLUT_LEFT_BUTTON] && (modifiers == GLUT_ACTIVE_SHIFT))
		|| (mouse.buttonPressed[GLUT_RIGHT_BUTTON])) {
		camera.y += 0.02 * dy;
	}

	mouse.oldX = x;
	mouse.oldY = y;
	glutPostRedisplay();
}

void mouseButton(int button, int mstate, int x, int y) {
	modifiers = glutGetModifiers();
	mouse.oldX = x;
	mouse.oldY = y;
	mouse.buttonPressed[button] = (mstate == GLUT_DOWN);
}

void mouseWheel(int wheel, int direction, int x, int y) {
	camera.y += 0.2 * direction;
}

void reshape(GLsizei width, GLsizei height) {
	DAVE::DaveWindowManager::WindowStruct* win = (DAVE::DaveWindowManager::WindowStruct*) glutGetWindowData();

	if (win) {
		win->width = width;
		win->height = height;
	}

	glViewport(0, 0, (GLsizei)(width), (GLsizei)(height));
}

void visible(int vis) {
	if (vis == GLUT_VISIBLE) glutIdleFunc(update);
		else glutIdleFunc(NULL);
}

///////////////////////////////////////////////////////////////////////////////

void drawScene(float totalTime, float diffTime, WinUserStruct* wus) {
	glRotatef(camera.camAngle1, 1, 0, 0);
	glRotatef(camera.camAngle2, 0, 0, 1);

	// render a triangle
	glBegin(GL_TRIANGLES);
		glColor3f(1, 0, 0);
		glVertex3f(3, 0, -2);
		glColor3f(0, 1, 0);
		glVertex3f(-3, 0, -2);
		glColor3f(0, 0, 1);
		glVertex3f(0, 0, 1.5);
	glEnd();
}

///////////////////////////////////////////////////////////////////////////////

int main(int argc, char* argv[]) {
	// DAVE
	dave.parseCmdLine(argc, argv);
	dave.init(sizeof(UserSyncDataStruct));

	glutInit(&argc, argv);

	for (std::vector<DAVE::DaveWindowManager::WindowStruct>::iterator win=dave.windowManager->windows.begin(); win!=dave.windowManager->windows.end(); ++win) {
		createOpengleanDaveWindow(&*win, GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
		glutKeyboardFunc(keyboardFunc);
		//glutSpecialFunc(keyboardSpecialFunc);
		glutReshapeFunc(reshape);
		glutMotionFunc(mouseDrag);
		glutPassiveMotionFunc(mouseMove);
		glutMouseFunc(mouseButton);
		glutMouseWheelFunc(mouseWheel);
		glutIdleFunc(update);
		glutDisplayFunc(display);
	}

	init();

	glutMainLoop();
	return 0;
}
