#include <ctype.h> // for isalpha()
#include "DaveConfig.h"
#include <iostream>
#include <fstream>
#include <algorithm> // for replace()
#include <sstream>
#include <string.h>

DAVE::DaveConfig::DaveConfig() {
}

DAVE::DaveConfig::~DaveConfig() {
}

// for settings/overrides on command line
// everything is processed in order so config file settings
// may be overridden by following command line settings
// use -davecfg configFileName on the command line to invoke
// a call to processConfigFile()
void DAVE::DaveConfig::processCmdLineArguments(char argc, char* argv[]) {
	for (int i=1; i<argc; i++) {
		if (strcmp(argv[i], "-davecfg") == 0) {
			if (i++ < argc) {			
				processConfigFile(argv[i]);
			} else {
				std::cerr << "Error: no config file specified" << std::endl;
			}
		} else {
			processLine(argv[i]);
		}
	}
}

// calls processLine() for each line
bool DAVE::DaveConfig::processConfigFile(char* configFileName) {
	std::ifstream confFile(configFileName);
	if (!confFile.is_open()) {
		std::cerr << "Error loading config file " << configFileName << std::endl;
		return false;
	}

	// process lines
	int lineNumber = 0;
	std::string line;
	while (std::getline(confFile, line)) {
		lineNumber++;
		if (!line.empty()) {
			processLine(line);
		}
	}

	confFile.close();
	return true;
}

// first token is key, second value
// '=' are treated as whitespaces (so they can be used)
// whitespaces are ignored (before and after value)
// value containing whitespaces can be enclosed in "" (meaning " are ignored)
// key has to start with alphabetical value (ignored otherwise)
bool DAVE::DaveConfig::processLine(std::string line) {
	if (!line.size()) return false;

	// replace all '=' with ' '
	std::replace(line.begin(), line.end(), '=', ' ');

	std::stringstream lineSs(line);
	std::string key, value;
	lineSs >> key;

	// check if first token exists and starts with a letter
	if (!key.size() || !isalpha(key[0])) return false;

	lineSs >> value;

	// special case for " (for values containing white spaces)
	if (value.c_str()[0] == '\"') {
		std::string valuePart2;
		std::getline(lineSs, valuePart2, '\"');
		value = (value.c_str()+1) + valuePart2; // +1: skip first "

		if (value.length() && (value.c_str()[value.length()-1]=='\"')) // in case of single word
			value.erase(value.length()-1); // skip " at the end
	}

	if (value.size()) config[key] = value;
	return true;
}

std::string DAVE::DaveConfig::getString(char* key) {
	return config[key];
}

std::string DAVE::DaveConfig::getString(std::string key) {
	return config[key];
}

double DAVE::DaveConfig::getDouble(char* key) {
	return atof(config[key].c_str());
}

double DAVE::DaveConfig::getDouble(std::string key, double defaultValue /*= 0*/) {
//	return atof(config[key].c_str());

	std::map<std::string, std::string>::iterator i = config.find(key);
	if (i==config.end()) return defaultValue;
	return atof(i->second.c_str());
}

int DAVE::DaveConfig::getInt(char* key) {
	return atoi(config[key].c_str());
}

int DAVE::DaveConfig::getInt(std::string key, int defaultValue/* = 0*/) {
//	return atoi(config[key].c_str());

	std::map<std::string, std::string>::iterator i = config.find(key);
	if (i==config.end()) return defaultValue;
	return atoi(i->second.c_str());
}

bool DAVE::DaveConfig::hasKey(char* key) {
	std::map<std::string, std::string>::iterator i = config.find(key);
	return (i!=config.end());
}

bool DAVE::DaveConfig::hasKey(std::string key) {
	std::map<std::string, std::string>::iterator i = config.find(key);
	return (i!=config.end());
}

// for debugging purposes
void DAVE::DaveConfig::printAll() {
	for(std::map<std::string,std::string>::iterator i=config.begin(); i!=config.end(); i++) {
		std::cout << i->first << " = " << i->second << std::endl;
	}
}
