// implementation of DaveNetwork using sample code from PracticalSockets
// using multicast udp datagrams (supports only one multicast group at the moment)

// Windows: link to Ws2_32.lib

#ifdef _DEBUG
#include <crtdbg.h>
#endif

#ifdef WIN32
  #include <winsock2.h>   // for win socket API calls; use Ws2_32.lib
  #include <ws2tcpip.h>   // for win socket structs; use Ws2_32.lib
#endif // WIN32

#include "../DaveNetwork.h"

#include "PracticalSocket.h"  // for UDPSocket and SocketException
#include <iostream>           // for cout and cerr

#include <string.h>
#include <stdlib.h>

// multicast address range: 224.0.0.0 through 239.255.255.255
bool DAVE::isMulticastAddress(const char* address) {
	if (!address) return false;

	int firstNum = atoi(address);
	if ((firstNum < 224) || (firstNum > 239)) return false;
	return true;
}

struct ServerUDPStruct {
	UDPSocket* sock; // Socket class of PracticalSocket
	std::string address; // address of multicast group
	int port;
};

DAVE::DaveNetworkServerUDP::DaveNetworkServerUDP(bool multicast/* = false*/) {
	#ifdef _DEBUG
		_ASSERTE(_CrtCheckMemory());
	#endif

	this->multicast = multicast;
	prdata = new ServerUDPStruct();

	unsigned char multicastTTL = 1; // Default TTL
	((ServerUDPStruct*) prdata)->sock = NULL;
	try {
		((ServerUDPStruct*) prdata)->sock = new UDPSocket();
		if (multicast) {
			((ServerUDPStruct*) prdata)->sock->setMulticastTTL(multicastTTL);
		}
	} catch (SocketException &e) {
		std::cerr << e.what() << std::endl;
		#ifdef _WIN32
		std::cerr << "WSA error: " << WSAGetLastError() << std::endl;
		#endif
		exit(1);
	}

	#ifdef _DEBUG
		_ASSERTE(_CrtCheckMemory());
	#endif
}

DAVE::DaveNetworkServerUDP::~DaveNetworkServerUDP() {
	if (((ServerUDPStruct*) prdata)->sock) delete ((ServerUDPStruct*) prdata)->sock;
	delete (ServerUDPStruct*) prdata;
}

// address is only used for multicast group
void DAVE::DaveNetworkServerUDP::addClient(const char* address, int port) {
	// copy string in case pointer becomes invalid
	((ServerUDPStruct*) prdata)->address = address;
	((ServerUDPStruct*) prdata)->port = port;
}

void DAVE::DaveNetworkServerUDP::sendStatusToClients(void* data, int size) {
	// send data to the multicast group
//std::cout << "sending data to " << ((ServerUDPStruct*) prdata)->address << ":" << ((ServerUDPStruct*) prdata)->port << std::endl;
	//std::cout << "sending string: " << (char*) data << std::endl;
	((ServerUDPStruct*) prdata)->sock->sendTo(data, size, ((ServerUDPStruct*) prdata)->address, ((ServerUDPStruct*) prdata)->port);
}

/////////////////////////////////////////////////////////////////////////////

struct ClientUDPStruct {
	UDPSocket* sock;
	char* serverAddress;
};

DAVE::DaveNetworkClientUDP::DaveNetworkClientUDP(const char* address, int port) {
	#ifdef _DEBUG
		_ASSERTE(_CrtCheckMemory());
	#endif

	multicast = isMulticastAddress(address);
	prdata = new ClientUDPStruct();

	((ClientUDPStruct*) prdata)->sock = NULL;
	((ClientUDPStruct*) prdata)->serverAddress = NULL;

	#ifdef _DEBUG
		_ASSERTE(_CrtCheckMemory());
	#endif

	// copy string in case pointer becomes invalid
	((ClientUDPStruct*) prdata)->serverAddress = new char[strlen(address)+1];
	strcpy(((ClientUDPStruct*) prdata)->serverAddress, address);
	//std::cout << ((ClientUDPStruct*) prdata)->serverAddress << ":" << port << std::endl;
	try {
		((ClientUDPStruct*) prdata)->sock = new UDPSocket(port);
//		((ClientUDPStruct*) prdata)->sock->set_noblock();
		if (multicast) {
			((ClientUDPStruct*) prdata)->sock->joinGroup(address); // multicast address
			//std::cout << "joined multicast group" << std::endl;
		}
	} catch (SocketException &e) {
		cerr << e.what() << endl;
#ifdef _WIN32
		std::cerr << "WSA error: " << WSAGetLastError() << std::endl;
#endif
		exit(1);
	}

	#ifdef _DEBUG
		_ASSERTE(_CrtCheckMemory());
	#endif
}

DAVE::DaveNetworkClientUDP::~DaveNetworkClientUDP() {
	if (((ClientUDPStruct*) prdata)->sock) delete ((ClientUDPStruct*) prdata)->sock;
	if (((ClientUDPStruct*) prdata)->serverAddress) delete ((ClientUDPStruct*) prdata)->serverAddress;
	delete (ClientUDPStruct*) prdata;
}

// returns size of received data (max. bufferSize bytes)
// data points to memory (... bytes) and is filled with the received bytes
// if something is received call this several times in a loop to prevent buffer overrun
// if nothing is received the method may block for a few milliseconds while waiting for a message/time out
int DAVE::DaveNetworkClientUDP::receiveStatusFromServer(void* data, int bufferSize) {
	try {
		unsigned short sourcePort;        // Port of datagram source
		std::string clAddr = ((ClientUDPStruct*) prdata)->serverAddress;
		//std::cout << "receiving ... ";
		// do not block and wait for message to arrive but return 0
		if (!((ClientUDPStruct*) prdata)->sock->hasData()) return 0;

		int bytesRcvd = ((ClientUDPStruct*) prdata)->sock->recvFrom(data, bufferSize, clAddr, sourcePort);
//		cout << "Received " << recvString << " from " << sourceAddress << ": " 
//			<< sourcePort << endl;
		return bytesRcvd;
	} catch (SocketException &e) {
		cerr << e.what() << endl;
#ifdef _WIN32
		std::cerr << "WSA error: " << WSAGetLastError() << std::endl;
#endif
		exit(1);
	}

	return 0;
}
