#include <math.h>
#include <GL/openglean.h> //?
#include "DaveProjection.h"
#include <fstream>
#include <iostream>
#include <stdlib.h>

#ifndef M_PI
#define M_PI 3.1415926
#endif // M_PI

DAVE::DaveProjection::DaveProjection(double lowerLeftX, double lowerLeftY, double lowerLeftZ,
			double lowerRightX, double lowerRightY, double lowerRightZ,
			double upperRightX, double upperRightY, double upperRightZ,
			double upperLeftX, double upperLeftY, double upperLeftZ) {

	// 4 corners of rectangular image area in world coordinate system
	// all points must lie in a plane and describe a rectangle
	worldCS_lowerLeftCorner[0] = lowerLeftX;
	worldCS_lowerLeftCorner[1] = lowerLeftY;
	worldCS_lowerLeftCorner[2] = lowerLeftZ;
	worldCS_lowerRightCorner[0] = lowerRightX;
	worldCS_lowerRightCorner[1] = lowerRightY;
	worldCS_lowerRightCorner[2] = lowerRightZ;
	worldCS_upperRightCorner[0] = upperRightX;
	worldCS_upperRightCorner[1] = upperRightY;
	worldCS_upperRightCorner[2] = upperRightZ;
	worldCS_upperLeftCorner[0] = upperLeftX;
	worldCS_upperLeftCorner[1] = upperLeftY;
	worldCS_upperLeftCorner[2] = upperLeftZ;

	// set default values
	zNear = 0.03;
	zFar = 300;

	hmdHFov = -1;

	// non-unit axis directions of screen in world coordinate system
	double worldCS_right[3];
	for (int i=0; i<3; i++) worldCS_right[i] = worldCS_lowerRightCorner[i]-worldCS_lowerLeftCorner[i];
	double worldCS_up[3];
	for (int i=0; i<3; i++) worldCS_up[i] = worldCS_upperLeftCorner[i]-worldCS_lowerLeftCorner[i];
	double worldCS_normal[3] = {worldCS_right[1]*worldCS_up[2] - worldCS_right[2]*worldCS_up[1],
		worldCS_right[2]*worldCS_up[0] - worldCS_right[0]*worldCS_up[2],
		worldCS_right[0]*worldCS_up[1] - worldCS_right[1]*worldCS_up[0]}; // cross product

	// unit axes of screen (in world coordinate system)
	double rightLength = sqrt(worldCS_right[0]*worldCS_right[0] + worldCS_right[1]*worldCS_right[1] + worldCS_right[2]*worldCS_right[2]);
	for (int i=0; i<3; i++) worldCS_unitRight[i] = worldCS_right[i]/rightLength;
	double upLength = sqrt(worldCS_up[0]*worldCS_up[0] + worldCS_up[1]*worldCS_up[1] + worldCS_up[2]*worldCS_up[2]);
	for (int i=0; i<3; i++) worldCS_unitUp[i] = worldCS_up[i]/upLength;
	double normalLength = sqrt(worldCS_normal[0]*worldCS_normal[0] + worldCS_normal[1]*worldCS_normal[1] + worldCS_normal[2]*worldCS_normal[2]);
	for (int i=0; i<3; i++) worldCS_unitNormal[i] = worldCS_normal[i]/normalLength;

	// calibration Matrix
	calibMatrix = NULL;
}

// for HMD use this constructor
DAVE::DaveProjection::DaveProjection(float hmdHFov, float hmdAspectRatio) {
	this->hmdHFov = hmdHFov;
	this->hmdAspectRatio = hmdAspectRatio;

	// set default values
	zNear = 0.03;
	zFar = 300;

	// calibration Matrix
	calibMatrix = NULL;
}

DAVE::DaveProjection::~DaveProjection() {
	if (calibMatrix) delete[] calibMatrix;
	calibMatrix = NULL;
}

bool DAVE::DaveProjection::loadCalibMatrix(const char* fileName) {
	if (calibMatrix) delete[] calibMatrix;
	calibMatrix = new double[16];

	std::ifstream indata;
	indata.open(fileName); // opens the file
	int i=0;
	if (indata) { // file could be opened
		while (!indata.eof() && (i < 16)) {
			indata >> calibMatrix[i++]; // read calibMatrix
		}
		indata.close();
	}

	if (i!=16) {
		std::cerr << "Error reading calibration matrix from file: " << fileName << std::endl;
		if (calibMatrix) delete[] calibMatrix;
		calibMatrix = NULL;
		return false;
	}

	// print matrix
	double* cm = calibMatrix;
	std::cout << "loaded calibration matrix:\n" << std::endl;
	for (int i=0; i<16; i+=4) {
		std::cout << calibMatrix[i+0] << " " << calibMatrix[i+1] << " " << calibMatrix[i+2] << " " << calibMatrix[i+3] << std::endl;
	}

	return true;
}

void DAVE::DaveProjection::setProjectionMatrix(float eyeX, float eyeY, float eyeZ, float* head) {
	// eye/camera position in world coordinate system
	double worldCS_eye[3] = {eyeX, eyeY, eyeZ};

	if (hmdHFov <= 0) { // projection wall
		double screenCS_distance = 0; // signed distance from eye to screen plane
		for (int i=0; i<3; i++) screenCS_distance += (worldCS_eye[i]-worldCS_lowerLeftCorner[i]) * worldCS_unitNormal[i];
	//	double worldCS_closestPoint[3]; // closest point from eye position on screen plane in world coordinate system
	//	for (int i=0; i<3; i++) worldCS_closestPoint[i] = worldCS_eye[i] - screenCS_distance*worldCS_unitNormal[i];

		double screenCS_right = 0;
		for (int i=0; i<3; i++) screenCS_right += (worldCS_lowerRightCorner[i]-worldCS_eye[i]) * worldCS_unitRight[i];
		double screenCS_left = 0;
		for (int i=0; i<3; i++) screenCS_left += (worldCS_lowerLeftCorner[i]-worldCS_eye[i]) * worldCS_unitRight[i];
		double screenCS_top = 0;
		for (int i=0; i<3; i++) screenCS_top += (worldCS_upperLeftCorner[i]-worldCS_eye[i]) * worldCS_unitUp[i];
		double screenCS_bottom = 0;
		for (int i=0; i<3; i++) screenCS_bottom += (worldCS_lowerLeftCorner[i]-worldCS_eye[i]) * worldCS_unitUp[i];

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();

		if (calibMatrix) {
			glMultMatrixd(calibMatrix);

		//std::cout << "cal: "; for (int imt=0; imt<16; imt+=4) std::cout << calibMatrix[imt+0] << " " << calibMatrix[imt+1] << " " << calibMatrix[imt+2] << " " << calibMatrix[imt+3] << std::endl;

		}

		// setup frustum in screen coordinate system
		double perspectiveFrustumMatrix[16]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	/*	perspectiveFrustumMatrix[0] = -2*screenCS_distance / (screenCS_right-screenCS_left);
		perspectiveFrustumMatrix[5] = -2*screenCS_distance / (screenCS_top-screenCS_bottom);
		perspectiveFrustumMatrix[8] = zNear/screenCS_distance * (screenCS_right+screenCS_left) / (screenCS_right-screenCS_left);
		perspectiveFrustumMatrix[9] = zNear/screenCS_distance * (screenCS_top+screenCS_bottom) / (screenCS_top-screenCS_bottom);
		perspectiveFrustumMatrix[10] = (zNear+zFar) / (zNear-zFar);
		perspectiveFrustumMatrix[11] = -1;
		perspectiveFrustumMatrix[14] = -2*zNear*zFar / (zNear-zFar);
	*/
		perspectiveFrustumMatrix[0] = 2*screenCS_distance / (screenCS_right-screenCS_left);
		perspectiveFrustumMatrix[5] = 2*screenCS_distance / (screenCS_top-screenCS_bottom);
		perspectiveFrustumMatrix[8] = (screenCS_right+screenCS_left) / (screenCS_right-screenCS_left);
		perspectiveFrustumMatrix[9] = (screenCS_top+screenCS_bottom) / (screenCS_top-screenCS_bottom);
		perspectiveFrustumMatrix[10] = (zNear+zFar) / (zNear-zFar);
		perspectiveFrustumMatrix[11] = -1;
		perspectiveFrustumMatrix[14] = 2*zNear*zFar / (zNear-zFar);

		//std::cout << "per: "; for (int imt=0; imt<16; imt+=4) std::cout << perspectiveFrustumMatrix[imt+0] << " " << perspectiveFrustumMatrix[imt+1] << " " << perspectiveFrustumMatrix[imt+2] << " " << perspectiveFrustumMatrix[imt+3] << std::endl;

		glMultMatrixd(perspectiveFrustumMatrix);

		// rotate everything
		double rotMatrix[16]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1};
		for (int i=0; i<3; i++) rotMatrix[i*4] = worldCS_unitRight[i];
		for (int i=0; i<3; i++) rotMatrix[i*4+1] = worldCS_unitUp[i];
		for (int i=0; i<3; i++) rotMatrix[i*4+2] = worldCS_unitNormal[i];
		glMultMatrixd(rotMatrix);
	} else {
		// HMD
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();

		if (calibMatrix) {
			glMultMatrixd(calibMatrix);
		}

		// standard perspective matrix
		double hw = tan(hmdHFov*M_PI/360); // half width
		double hh = hw/hmdAspectRatio; // half height
		double perspectiveFrustumMatrix[16]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
		perspectiveFrustumMatrix[0] = 1/hw;
		perspectiveFrustumMatrix[5] = 1/hh;
		perspectiveFrustumMatrix[10] = (zNear+zFar) / (zNear-zFar);
		perspectiveFrustumMatrix[11] = -1;
		perspectiveFrustumMatrix[14] = 2*zNear*zFar / (zNear-zFar);
/*		// standard perspective matrix, looking in y direction
		double hw = tan(hmdHFov*M_PI/360); // half width
		double hh = hmdAspectRatio * hw; // half height
		double perspectiveFrustumMatrix[16]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
		perspectiveFrustumMatrix[0] = 1/hw;
		perspectiveFrustumMatrix[6] = 1/hh;
		perspectiveFrustumMatrix[9] = (zNear+zFar) / (zNear-zFar);
		perspectiveFrustumMatrix[11] = -1;
		perspectiveFrustumMatrix[13] = 2*zNear*zFar / (zNear-zFar);
*/		//std::cout << "per: "; for (int imt=0; imt<16; imt+=4) std::cout << perspectiveFrustumMatrix[imt+0] << " " << perspectiveFrustumMatrix[imt+1] << " " << perspectiveFrustumMatrix[imt+2] << " " << perspectiveFrustumMatrix[imt+3] << std::endl;
		glMultMatrixd(perspectiveFrustumMatrix);
		glRotatef(-90, 1, 0, 0);
		// rotate everything
		double rotMatrix[16]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1};
if (!head) {
	std::cout << "Error: Head rotation not transmitted to clients yet! HMDs work on master only.\n" << std::endl;
	exit(0);
}
		for (int i=0; i<3; i++) for (int j=0; j<3; j++) rotMatrix[i*4+j] = head[j*4+i];
		glMultMatrixd(rotMatrix);
	}


	// move to eye/camera position
	glTranslated(-worldCS_eye[0], -worldCS_eye[1], -worldCS_eye[2]);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}
