#include "DaveWindowManager.h"
#include "DaveProjection.h"
#include "DaveConfig.h"

#include <iostream> // for cerr

#ifdef WITH_BLENDMASK
#include "DaveBlendMask_PGM/PgmImage.h"
//#include "DaveBlendMask_YImage/YImage.hpp"
#endif // WITH_BLENDMASK

// from gl.h
#ifndef GL_BACK_LEFT
#define GL_BACK_LEFT   0x0402
#endif
#ifndef GL_BACK_RIGHT
#define GL_BACK_RIGHT  0x0403
#endif


DAVE::DaveWindowManager::DaveWindowManager(DAVE::DaveConfig* config) {
	this->config = config;
	iName = config->getString("InstanceName");

	WindowStruct w;
	DBufferStruct db;

	for (int winNr=1; winNr<=9; winNr++) {
		bool defaultWin = false;
		std::string prefix = iName + ".win" + ((char) (winNr+'0')) + ".";

		if (config->hasKey(prefix + "windowPosX") ||
			config->hasKey(prefix + "windowPosY") ||
			(defaultWin = ((windows.size() == 0) && (winNr == 9))) ) { // if we have no window at winNr==9, create one default window

			double lowerLeftX = winGetDouble("lowerLeftX", -0.2, winNr);
			double lowerLeftY = winGetDouble("lowerLeftY", 0.5, winNr);
			double lowerLeftZ = winGetDouble("lowerLeftZ", -0.15, winNr);
			double lowerRightX = winGetDouble("lowerRightX", 0.2, winNr);
			double lowerRightY = winGetDouble("lowerRightY", 0.5, winNr);
			double lowerRightZ = winGetDouble("lowerRightZ", -0.15, winNr);
			double upperRightX = winGetDouble("upperRightX", 0.2, winNr);
			double upperRightY = winGetDouble("upperRightY", 0.5, winNr);
			double upperRightZ = winGetDouble("upperRightZ", 0.15, winNr);
			double upperLeftX = winGetDouble("upperLeftX", -0.2, winNr);
			double upperLeftY = winGetDouble("upperLeftY", 0.5, winNr);
			double upperLeftZ = winGetDouble("upperLeftZ", 0.15, winNr);

			double hmdHFov = winGetDouble("hmdHFov", -1, winNr);

			w.posX = winGetInt("windowPosX", 10, winNr);
			w.posY = winGetInt("windowPosY", 10, winNr);
			w.width = winGetInt("windowWidth", 400, winNr);
			w.height = winGetInt("windowHeight", 300, winNr);
			w.fullScreenAtStartup = winGetString("windowMode", "", winNr) == "FULLSCREEN";
			std::string defaultTitle = "DAVE application";
			if (!defaultWin) defaultTitle += std::string(" (") + ((char) (winNr+'0')) + ")";
			w.title = winGetString("title", defaultTitle, winNr).c_str();

			// optional parameter, by default square pixels are assumed
			double hmdAspectRatio = winGetDouble("hmdAspectRatio", w.width/(double)w.height, winNr);

			if (hmdHFov > 0) {
				w.projection = new DAVE::DaveProjection(
					hmdHFov, hmdAspectRatio);
			} else {
				w.projection = new DAVE::DaveProjection(
					lowerLeftX, lowerLeftY, lowerLeftZ,
					lowerRightX, lowerRightY, lowerRightZ,
					upperRightX, upperRightY, upperRightZ,
					upperLeftX, upperLeftY, upperLeftZ);
			}

			// calibration matrix
			std::string cm = winGetString("calibMatrix", "", winNr);
			if (!cm.empty()) {
				w.projection->loadCalibMatrix(cm.c_str());
			}

			w.blendMaskImg = NULL;
			#ifdef WITH_BLENDMASK
			// soft edge blending mask image
			std::string blendMaskImgName = winGetString("blendMask", "", winNr);
			if (!blendMaskImgName.empty()) {
				//YImage* blendMaskImgYImg = new YImage();
				PgmImage* blendMaskImgPgmImg = new PgmImage();
				//if (!blendMaskImgYImg->load(blendMaskImgName.c_str())) {
				if (!blendMaskImgPgmImg->load(blendMaskImgName)) {
					std::cerr << "Error: couldn't open blendMask " << blendMaskImgName << std::endl;
				}
				//w.blendMaskImg = (void*) blendMaskImgYImg;
				w.blendMaskImg = (void*) blendMaskImgPgmImg;
			}
			#endif // WITH_BLENDMASK

			// clipping plane distances override
			double zn = winGetDouble("zNear", 0, winNr);
			if (zn > 0) w.projection->zNear = zn;
			double zf = winGetDouble("zFar", 0, winNr);
			if (zf > 0) w.projection->zFar = zf;

			db.windowIndex = windows.size(); // before push_back

			w.borderLess = false;
			if (winGetString("windowDecoration", "", winNr) == "BORDERLESS") w.borderLess = true;

			db.cameraStereoOffset = DBufferStruct::DEFAULT;
			std::string camStOffsStr = winGetString("cameraStereoOffset", "", winNr);
			if (camStOffsStr == "LEFT_EYE") db.cameraStereoOffset = DBufferStruct::LEFT_EYE;
			if (camStOffsStr == "RIGHT_EYE") db.cameraStereoOffset = DBufferStruct::RIGHT_EYE;

			db.quadDrawBuffer = 0;
			w.quadBuffer = false;

			bool quadBufferReverse = false;
			if (winGetString("stereoMode", "", winNr) == "STEREO_QUAD_BUFFER") w.quadBuffer = true;
			if (winGetString("stereoMode", "", winNr) == "STEREO_QUAD_BUFFER_REV") {
				w.quadBuffer = true;
				quadBufferReverse = true;
			}

			w.winUserDataPtr = NULL;

			windows.push_back(w);
			if (!w.quadBuffer) {
				// normal window with one double buffer
				dBuffers.push_back(db);
			} else {
				// two double buffers; set correct drawing buffer and set cameraStereoOffset
				db.quadDrawBuffer = GL_BACK_LEFT;
				db.cameraStereoOffset = quadBufferReverse ? DBufferStruct::RIGHT_EYE : DBufferStruct::LEFT_EYE;
				dBuffers.push_back(db);

				db.quadDrawBuffer = GL_BACK_RIGHT;
				db.cameraStereoOffset = quadBufferReverse ? DBufferStruct::LEFT_EYE : DBufferStruct::RIGHT_EYE;
				dBuffers.push_back(db);
			}
		}
	}
}

DAVE::DaveWindowManager::~DaveWindowManager() {
	#ifdef WITH_BLENDMASK
	for (std::vector<WindowStruct>::iterator w=windows.begin(); w!=windows.end(); ++w) {
		if (w->blendMaskImg) {
			//YImage* blendMaskImg = (YImage*) w->blendMaskImg;
			PgmImage* blendMaskImgPgmImg = (PgmImage*) w->blendMaskImg;
			delete blendMaskImgPgmImg;
			w->blendMaskImg = NULL;
		}
	}
	#endif // WITH_BLENDMASK
}

// example: winGetDouble('lowerLeftX', -0.2, 1) looks in 'instance.win1.lowerLeftX', then 'instance.lowerLeftX', then 'lowerLeftX', then -0.2
double DAVE::DaveWindowManager::winGetDouble(char* key, double defaultValue, int winNr /*= 0*/) {
	double retVal = config->getDouble(key, defaultValue);
	retVal = config->getDouble(iName + "." + key, retVal);
	if ((winNr >= 1) && (winNr <= 9)) {
		retVal = config->getDouble(iName + ".win" + ((char) (winNr+'0')) + "." + key, retVal);
	}
	return retVal;
}

int DAVE::DaveWindowManager::winGetInt(char* key, int defaultValue, int winNr /*= 0*/) {
	int retVal = config->getInt(key, defaultValue);
	retVal = config->getInt(iName + "." + key, retVal);
	if ((winNr >= 1) && (winNr <= 9)) {
		retVal = config->getInt(iName + ".win" + ((char) (winNr+'0')) + "." + key, retVal);
	}
	return retVal;
}

std::string DAVE::DaveWindowManager::winGetString(char* key, std::string defaultValue, int winNr /*= 0*/) {
	std::string retVal = config->getString(key);
	if (!retVal.empty()) return retVal;

	retVal = config->getString(iName + "." + key);
	if (!retVal.empty()) return retVal;

	if ((winNr >= 1) && (winNr <= 9)) {
		retVal = config->getString(iName + ".win" + ((char) (winNr+'0')) + "." + key);
		if (!retVal.empty()) return retVal;
	}

	return defaultValue;
}
